'<TiCoBasic Header, Headerversion 001.001>
' Foldings                       = 29,41,118
'<Header End>
'##############################################################################
'#                        CAN_TiCo.inc                                        #
'#              INCLUDE-file for CAN with Tico                                #
'##############################################################################
'#                                                                            #
'#                                                                            #
'#  19.09.2007  mf   creation                                                 #
'#  27.11.2007 mf bug get line status                                         #
'#  01.02.2008 mf #If adwin_system                                            #
'#  19.02.2009  MF/OL Read_Msg_Con                                            #
'#  19.02.2009  MF Transmit_Status                                            #
'#  24.03.2009  MF '''+                                                       #
'#  06.04.2010  MF (channel)(msg_no)                                          #
'#  01.06.2010  MF optimize init_can                                          #
'#  27.06.2012  mf can_set_mode                                               #
'#  30.04.2014  gc   inserted regions                                         #
'#  04.11.2016  mf Calc_Processdelay                                          #
'#  25.01.2021  mf   Calc_TicksToNs                                           #
'#                                                                            #
'# Attention:                                                                 #
'#   - Do not copy the first source line (containing                          #
'#     "'ADBASIC-SYSTEM-FILE...") to user files !                             #
'#     Otherwise the Syntax color may be not correct.                         #
'#   - This includefile requires ADbasic 4.20.04 or higher!                   #
'#                                                                            #
'#  Eventsource with clear (3001H) without clear (3000H)                      #
'##############################################################################

'<REGION> Variables
#If MAKE_LIB = OFF Then
Dim PWM_Help_mode as Long ''<HIDE>
Dim CAN_msg[9] As Long at dm_local   ' array with 9 elements
#EndIf
'<ENDREGION>

#If adwin_system=ADwin_ProII
#Else
#error("This include file can only be used with the ADwin ProII")
#EndIf

'<REGION> System

'***************************************************************
''' Converts a process frequency [Hz] into processor ticks (processdelay or cycle time).
''' frequency:  frequency[Hz]
'''*************************************************************
Function Calc_Processdelay(frequency) as long
  
#if prozessor=tico1
  calc_processdelay=50000000/(frequency)
#else
  calc_processdelay=100000000/(frequency)
#endif
  
endfunction




'***************************************************************
''' Converts timer ticks into ns
''' ticks: counter value ticks into ns
'''*************************************************************
Function Calc_TicksToNs(ticks) As long
  Calc_TicksToNs=(ticks)*20
endfunction


'---------------------------------------------------------------
' If a low-priority process accesses must not be
' interrupted by any other process !!!
'---------------------------------------------------------------

Sub Sub_Begin_Critical_Section() ''<HIDE>
#If PRIORITAET = LOW
  Begin_Critical_Section
#EndIf
EndSub

Sub Sub_End_Critical_Section() ''<HIDE>
#If PRIORITAET = LOW
  End_Critical_Section
#EndIf
EndSub

'******************************************************************************
'''Switches the LED on or off.
'''value: 
'''+   0 off
'''+   1 on
'''*****************************************************************************
Sub Set_LED(value)
  out (80804000H, value)
EndSub

'******************************************************************************
'''Sets CAN LEDs on/off for the CAN channel.
'''channel: number of the CAN interface on the CAN module
'''+(also necessary If the module provides only one channel)
'''led_col: 
'''+         0 = LED off 
'''+         1 = LED red  
'''+         2 =  LED green 
'''+         3 = LED orange
'''****************************************************************************
Sub CAN_Set_LED(channel, led_col)
  out (10000150H+((channel)*1000h),led_col)
EndSub

'****************************************************************************
'''Returns the status of the LED (on top of the front panel) of the module.
'''**************************************************************************
Function Check_LED() AS LONG
  Check_LED=in(80804000h)
EndFunction
'<ENDREGION>

'<REGION> CAN
'****************************************************************************
'''Sets CAN Mode. 
'''channel: (1..2)
'''mode:
'''+ 0 OFF    (RB=OFF)
'''+ 1 OFF    (RB=120 OHM)
'''+ 2 HS CAN (RB=OFF)
'''+ 3 HS CAN (RB=120 OHM)
'''+ 4 LS CAN (RB=OFF) 
'''*****************************************************************************
Sub CAN_Set_Mode(channel,mode)
  dim md as long
  dim rb as long  
  dim old as long
  
  rb=0
  if ((mode=1) or (mode=3)) then rb=1
    
  md=rb 'off
  if ((mode=2)  or (mode=3)) then md=md or 2 'high
  if (mode=4) then md=md or 4 'low
  
  old=in(10000140h OR (1000h*(channel)))
    
  if (old<>md) then  
    out(10000140h OR (1000h*(channel)),rb)  'set rb /driver off
    Sleep(100000) '10 ms relais settling time
    out(10000140h OR (1000h*(channel)),md) 'driver on
    Sleep(100000) '10 ms relais settling time
  endif
  
ENDSUB

'****************************************************************************
'''Enables a message object for receiving messages.
'''+Disables all interrups of the message object.
'''channel: number of the CAN interface on the CAN module (also necessary If the module provides only one channel)
'''msg_no:  Number of the message object
'''identifier: incoming frames are accepted by this message object,If the identifier matches the frame identifier
'''id_length: Determines whether the identifier is 11 or 29 Bit
'''+             0 = 11 Bit identifier
'''+             1 = 29 Bit identifier
'''*****************************************************************************
Sub En_Receive(channel, msg_no, identifier, id_length)
  Set_CAN_Reg(channel,msg_no * 16,149)                                       'enable message object
  If (id_length = 0) Then
    Set_CAN_Reg(channel,(msg_no) * 16 + 6,128)                                 'read; DLC = 8 ; 11 Bit identifier
    Set_CAN_Reg(channel,(msg_no) * 16 + 2,(Shift_Right(identifier,3) And 255))  'Set identifier
    Set_CAN_Reg(channel,(msg_no) * 16 + 3,(Shift_Left(identifier,5) And 255))
  Else
    Set_CAN_Reg(channel,(msg_no) * 16 + 6,132)                                 'read; DLC = 8 ; 29 Bit identifier
    Set_CAN_Reg(channel,(msg_no) * 16 + 2,(Shift_Right(identifier,21) And 255)) 'Set identifier
    Set_CAN_Reg(channel,(msg_no) * 16 + 3,(Shift_Right(identifier,13) And 255))
    Set_CAN_Reg(channel,(msg_no) * 16 + 4,(Shift_Right(identifier,5) And 255))
    Set_CAN_Reg(channel,(msg_no) * 16 + 5,(Shift_Left(identifier,3) And 255))
  EndIf
EndSub

'****************************************************************************
'''Enables a message object for transmitting messages.
'''+Disables all interrups of the message object.
'''channel: number of the CAN interface on the CAN module (also necessary If the module provides only one channel)
'''msg_no:  Number of the message object
'''identifier: identifier for the frames which will be transmitted
'''id_length: Determines whether the identifier is 11 or 29 Bit
'''+             0 = 11 Bit identifier
'''+             1 = 29 Bit identifier
'''*****************************************************************************
Sub En_Transmit(channel, msg_no, identifier, id_length)
  Dim bit As Long
  Set_CAN_Reg(channel,msg_no * 16,149)                                       'enable message object
  If ((id_length And 2) = 0) Then
    bit = 0
  Else
    bit = 8
  EndIf
  If ((id_length And 1) = 0) Then
    Set_CAN_Reg(channel,(msg_no) * 16 + 6,(136 - bit))                          'write; DLC = 8 ; 11 bit identifier
    Set_CAN_Reg(channel,(msg_no) * 16 + 2,(Shift_Right(identifier,3) And 255))   'Set identifier
    Set_CAN_Reg(channel,(msg_no) * 16 + 3,(Shift_Left(identifier,5) And 255))
  Else
    Set_CAN_Reg(channel,(msg_no) * 16 + 6,(140 - bit))                          'write; DLC = 8 ; 29 bit identifier
    Set_CAN_Reg(channel,(msg_no) * 16 + 2,(Shift_Right(identifier,21) And 255))  'Set identifier
    Set_CAN_Reg(channel,(msg_no) * 16 + 3,(Shift_Right(identifier,13) And 255))
    Set_CAN_Reg(channel,(msg_no) * 16 + 4,(Shift_Right(identifier,5) And 255))
    Set_CAN_Reg(channel,(msg_no) * 16 + 5,(Shift_Left(identifier,3) And 255))
  EndIf
EndSub

'****************************************************************************
'''Initializes the CAN-controller.
'''+All message objects will be disabled.
'''+Bit rate: 1 MBit/s
'''+Acceptance filter: Must match
'''channel: number of the CAN interface on the CAN module (also necessary If the module provides only one channel)
'''*****************************************************************************
Sub Init_CAN(channel)
  Dim counter As Long

  out((10000000h OR ((1000h * (channel)) or 120H)),1)
  Sleep(12000)                                            ' Reset pin must be low for at least 1 ms

  Set_CAN_Reg(channel,0,65)                           'Access to the registers of the controller only for CPU
  Set_CAN_Reg(channel,2,65)                           'CPU interface register
 
  for counter=6 to 15
    Set_CAN_Reg(channel,counter,0FFh)                       'Setting all masks to "must match"
  next   
  'Set_CAN_Reg(channel,06h,0FFh)                       'Setting all masks to "must match"
  'Set_CAN_Reg(channel,07h,0FFh)
  'Set_CAN_Reg(channel,08h,0FFh)
  'Set_CAN_Reg(channel,09h,0FFh)
  'Set_CAN_Reg(channel,0Ah,0FFh)
  'Set_CAN_Reg(channel,0Bh,0FFh)
  'Set_CAN_Reg(channel,0Ch,0FFh)
  'Set_CAN_Reg(channel,0Dh,0FFh)
  'Set_CAN_Reg(channel,0Eh,0FFh)
  'Set_CAN_Reg(channel,0Fh,0FFh)
  Set_CAN_Reg(channel,01Fh,0)                         ' Clock Out Register
  Set_CAN_Reg(channel,02Fh,0)                         ' Bus configuration register
  Set_CAN_Reg(channel,03Fh,0)                         ' Bit Timing register 0
  Set_CAN_Reg(channel,04Fh,032h)                      ' Bit Timing register 1

  For counter = 1 To 15                                       'Disable all message objects
    Set_CAN_Reg(channel,counter * 16,85)
    Set_CAN_Reg(channel,counter * 16 + 1,85)
  Next counter
  Set_CAN_Reg(channel,0,0)                            'Access to the registers of the controller only for Application
EndSub

'---------------------------------------------------------------
' Aid Sub for Read_Msg_Con and Read_Msg.
'---------------------------------------------------------------
Sub Read_Controller_Registers(channel, msg_no, id0, id1, id2, id3, id_length)''<HIDE>
  Dim counter As Long
  Dim buffer As Long

  buffer = (Get_CAN_Reg(channel,(msg_no) * 16 + 6))
  CAN_msg[9] = Shift_Right((buffer And 0F0h),4)             ' Number of databytes
  id_length = buffer And 4
  id0 = Get_CAN_Reg(channel,(msg_no) * 16 + 2)         ' identifier registers
  id1 = Get_CAN_Reg(channel,(msg_no) * 16 + 3)
  If (id_length = 4) Then
    id2 = Get_CAN_Reg(channel,(msg_no) * 16 + 4)
    id3 = Get_CAN_Reg(channel,(msg_no) * 16 + 5)
  EndIf
  For counter = 1 To CAN_msg[9]                             ' reading databytes
    CAN_msg[counter] = Get_CAN_Reg(channel,((msg_no) * 16)  + counter + 6)
  Next counter
EndSub

'****************************************************************************
'''Checks if there is a new frame in the selected
'''+message object. If there is a new frame all databytes are read
'''+and then written to the array "CAN_msg".
'''+The number of databytes is written to the element 9 of the array "CAN_msg".
'''+The return value of the Function is the identifier of the frame.
'''+If there is no new frame in the message box the return value is -1.
'''channel: number of the CAN interface on the CAN module (also necessary If the module provides only one channel)
'''msg_no: number of the message object
'''*****************************************************************************
Function Read_Msg(channel, msg_no) As Long
  Dim new As Long                                         ' new message in the message object
  Dim id0,id1,id2,id3 As Long                             ' identifier Byte 0 and 1
  Dim id_length As Long                                   ' buffer to distinguish 11 and 29 bit identifier

  new = Get_CAN_Reg(channel,(msg_no) * 16 + 1) And 3        ' Check,If a new frame in arrived
  If (new = 2) Then                                       ' New frame is arrived
    Read_Controller_Registers(channel,msg_no,id0,id1,id2,id3,id_length)
    Set_CAN_Reg(channel,(msg_no) * 16 ,0FDh)                ' reset interrupt
    Set_CAN_Reg(channel,(msg_no) * 16  + 1 ,85)             ' reset the new data flag
    If (id_length = 0) Then
      Read_Msg = (Shift_Right(id1,5)) Or (Shift_Left(id0,3))  ' return value = 11 bit identifier
    Else
      Read_Msg = (Shift_Right(id3,3)) Or (Shift_Left(id2,5)) Or (Shift_Left(id1,13)) Or (Shift_Left(id0,21))        ' return value = 29 bit identifier
    EndIf
  Else
    Read_Msg = -1                                         ' return value = 0
  EndIf
EndFunction

'****************************************************************************
'''Checks if there is a new frame in the selected
'''+message object. If there is a new frame all databytes are read
'''+and then written to the array "CAN_msg".
'''+The number of databytes is written to the element 9 of the array "CAN_msg".
'''+The return value of the Function is the identifier of the frame.
'''+If there is no new frame in the message box the return value is -1.
'''+Checks the consistens of the data and If they are not consistent
'''+the datas will be reread.
'''channel: number of the CAN interface on the CAN module (also necessary If the module provides only one channel)
'''msg_no: number of the message object
'''*****************************************************************************
Function Read_Msg_Con(channel, msg_no) As Long
  Dim new As Long                                         ' new message in the message object
  Dim id_length As Long                                   ' buffer to distinguish 11 and 29 bit identifier
  Dim id0,id1,id2,id3 As Long                             ' identifier Byte 0 and 1

  new = Get_CAN_Reg(channel,(msg_no) * 16 + 1) And 3        ' Check,If a new frame in arrived
  If (new = 2) Then                                       ' New frame is arrived
    If (msg_no = 15) then
      read_controller_registers(channel, msg_no, id0, id1, id2, id3, id_length)
      Set_CAN_Reg(channel, (msg_no) * 16, 0FDh)                                            ' reset interrupt
      Set_CAN_Reg(channel, (msg_no) * 16 + 1, 85)                                          ' reset the new data flag
      new = Get_CAN_Reg( channel, (msg_no) * 16 + 1) And 3                                  ' Check, if a new frame in arrived
      If (new = 2) Then                                                                             ' New frame is arrived
        read_controller_registers(channel, msg_no, id0, id1, id2, id3, id_length)
        Set_CAN_Reg( channel, (msg_no) * 16, 0FDh)                                          ' reset interrupt
        Set_CAN_Reg( channel, (msg_no) * 16 + 1, 85)                                        ' reset the new data flag
      endif
    else
      Set_CAN_Reg(channel,(msg_no) * 16  + 1 ,85)             ' reset the new data flag
      Read_Controller_Registers(channel,msg_no,id0,id1,id2,id3,id_length)
      new = Get_CAN_Reg(channel,(msg_no) * 16 + 1) And 3      ' Check,If a new frame in arrived
      If (new = 2) Then                                     ' New frame is arrived
        Read_Controller_Registers(channel,msg_no,id0,id1,id2,id3,id_length)
        Set_CAN_Reg(channel,(msg_no) * 16  + 1 ,85)           ' reset the new data flag
      EndIf
      Set_CAN_Reg(channel,(msg_no) * 16 ,0FDh)                ' reset interrupt
     
    endif
      
    If (id_length = 0) Then
      Read_Msg_Con = (Shift_Right(id1,5)) Or (Shift_Left(id0,3))  ' return value = 11 bit identifier
    Else
      Read_Msg_Con = (Shift_Right(id3,3)) Or (Shift_Left(id2,5)) Or (Shift_Left(id1,13)) Or (Shift_Left(id0,21))        ' return value = 29 bit identifier
    EndIf
  Else
    Read_Msg_Con = -1                                     ' return value = 0
  EndIf
EndFunction

'****************************************************************************
'''Transmits a frame.
'''+The databytes and the number of databytes have to be written in the
'''+array "CAN_msg" before this Subroutine is activated.
'''+After this Subroutine the frame is transmitted As soon As the
'''+CAN-bus allows it.
'''channel: number of the CAN interface on the CAN module
'''msg_no: number of the message object
'''*****************************************************************************
Sub Transmit(channel, msg_no)
  Dim buffer As Long
  Dim counter As Long

  For counter = 1 To CAN_msg[9]
    Set_CAN_Reg(channel,((msg_no) * 10h) + counter + 6,CAN_msg[counter])       'copy datas
  Next counter

  ' Anzahl der Datenbytes bergeben
  buffer = Get_CAN_Reg(channel,((msg_no) * 10h) + 6)      'read register
  buffer = buffer And 0Fh                                     'change DLC Bits
  buffer = buffer Or Shift_Left(CAN_msg[9],4)
  Set_CAN_Reg(channel,(msg_no)  * 16  + 6,buffer)          'write register
  Set_CAN_Reg(channel,(msg_no) * 16  + 1,102)           ' Control register (setting bit to transmit data)
EndSub

'****************************************************************************
'''Checks if a Messagebox is ready to send. (The last Message is
'''+already sent).
'''+Return: 
'''+        0 = ready to send
'''+        1 = not ready to send
'''channel: number of the CAN interface on the CAN module
'''msg_no: number of the message object
'''*****************************************************************************
Function Transmit_Status(channel, msg_no) as long
  Transmit_Status = shift_right(Get_CAN_Reg(channel,((msg_no) * 10h) + 1), 5) and 1    ' Control register
EndFunction

'****************************************************************************
'''Enables the interrupt of a message object.
'''channel: number of the CAN interface on the CAN module
'''msg_no: number of the message object
'''-------------------------------------------------------------
Sub En_Interrupt(channel, msg_no)
  Set_CAN_Reg(channel,(msg_no) * 16,0EBh)              'Enable interrupt of the message object.
  Set_CAN_Reg(channel,0,02h)                         'Enable global interrupt.
EndSub

'****************************************************************************
'''Sets a register of the CAN-controller.
'''channel: number of the CAN interface on the CAN module
'''regno: register which should be set
'''value: value which should be written to the register.
'''*****************************************************************************
Sub Set_CAN_Reg(channel, regno, value)
  Dim h as long

  Sub_Begin_Critical_Section
  out (10000000h OR (1000h * (channel)) or (regno and 0ffh),value)
  sleep(7) '
  Sub_End_Critical_Section

EndSub

'****************************************************************************
'''Reads a register of the CAN-controller.
'''+Return value: value of the register
'''channel: number of the CAN interface on the CAN module
'''+(also necessary if the module provides only one channel)
'''regno: register to read
'''*****************************************************************************
Function Get_CAN_Reg(channel, regno) As Long
  Dim old as long

  SUB_BEGIN_CRITICAL_SECTION
  old=in(10000000h OR (1000h * (channel)) or ((regno) and 0ffh))
  sleep(7)
  Get_CAN_Reg= in(10000000h OR (1000h * (channel)) or (100h))
  SUB_END_CRITICAL_SECTION

EndFunction

'****************************************************************************
'''Sets the baudrate of the CAN-Bus.
'''+return value 
'''+             0 = baudrate is set
'''+             1 = the selected baudrate is not set
'''channel: number of the CAN interface on the CAN module
'''+(also necessary If the module provides only one channel)
'''baudrate: CAN-Bus baudrate
'''*****************************************************************************
Function Set_CAN_Baudrate(channel, baudrate) As Long
  Dim scale As Long
  Dim fbuffer As long
  Dim lbuffer As Long
  Dim back As Long
  Dim tseg1 As Long
  Dim tseg2 As Long
  Dim brp As Long
  Dim btr0 As Long
  Dim btr1 As Long

  Dim hscale as long
  Dim hbaud as long

  fbuffer = 8000000 / baudrate

  scale =fbuffer 'total scale to achive the baudrate
  hbaud=(fbuffer)*baudrate

  If (absi(hbaud-8000000)<1100) then
    Set_CAN_Baudrate=0'                                '1 If baudrate not possible
  Else
    Set_CAN_Baudrate=1
  EndIf

  If (Set_CAN_Baudrate = 0) Then                         'If baudrate possible then find settings
    brp = (scale - 1) / 25                               'Prescaler
    Do                                                   'find the right brp
      fbuffer=scale/(brp+1)-3
      hscale=(fbuffer+3)*(brp+1)                        'Inc prescaler until tseg1 + tseg 2 is a Long
      back=scale-hscale
      Inc(brp)
    Until((back = 0) Or (fbuffer<5))
    Dec(brp)
    If (back <> 0) Then Set_CAN_Baudrate = 1
  EndIf

  If (Set_CAN_Baudrate = 0) Then                         'If there is a correct prescaler
    tseg2 = fbuffer * 4/10
    tseg1 = fbuffer - tseg2
    If (tseg2 > 7) Then
      Do
        Dec(tseg2)
        Inc(tseg1)
      Until(tseg2 = 7)
    EndIf
    btr0 = brp                                           'Synchronization jump width = 1 (bits = 0)
    btr1 = tseg1 + Shift_Left(tseg2,4)                   'one sampling each Bit
    Set_CAN_Reg(channel,0,65)                     'enable access to btr regisers
    Set_CAN_Reg(channel,3Fh,btr0)                 'set btr0 regisers
    Set_CAN_Reg(channel,4Fh,btr1)                 'set btr1 regisers
    Set_CAN_Reg(channel,0,0)                      'disable access to btr registers
  EndIf
EndFunction

'---------------------------------------------------------------
' Reading the register where the bus error pin of the tranceiver
' is stored.
' The register is set when the a bus error occurs and is reset by
' reading the register.
' module:  module address
' channel : number of the CAN interface on the CAN module
'          (also necessary If the module provides only one channel)
' return value : 0 = no error
'                1 = error
' Attention: This command is only for low-speed CAN
'---------------------------------------------------------------
Function Get_Line_Status(channel) As Long ''<HIDE>
  Dim old as long

  old=in(10000120h OR  (1000h * (channel)))
  Get_Line_Status=shift_right(old,1) AND 1

EndFunction

'<ENDREGION>
