'<TiCoBasic Header, Headerversion 001.001>
' Foldings                       = 28,32,113,244,360,786,1250,1485
'<Header End>
'##############################################################################
'#                        FIELDBUS_TiCo.inc                                   #
'#              INCLUDE-file for PRO II PROFIBUS with Tico                    #
'##############################################################################
'#                                                                            #
'#                                                                            #
'#  26.02.2008  mf   creation                                                 #
'#  20.03.2008  ur   PROFIBUS                                                 #
'#  13.05.2008  mf   ABCC_Reset()                                             #
'#  26.03.2009  ur   DeviceNet                                                #
'#  10.08.2009  UR   Profibus/DeviceNet process array element value           #
'#                   limit to 8bit                                            #
'#  26.01.2012  ur   Anybus error handling / error codes update               #
'#  30.04.2014  gc   constants hidden, regions                                #
'#  04.11.2016  mf   Calc_Processdelay                                        #
'#  12.12.2018  mf   AnyBus M40                                               #
'#  12.12.2018  ol   Profinet,Profibus, Ethercat M 40 eingefgt               #
'#  25.01.2021  mf   Calc_TicksToNs                                           #
'#                                                                            #
'# Attention:                                                                 #
'#   - Do not copy the first source line (containing                          #
'#     "'ADBASIC-SYSTEM-FILE...") to user files !                             #
'#     Otherwise the Syntax color may be not correct.                         #
'#                                                                            #
'##############################################################################

#if ADwin_System = ADwin_ProII
#else
#error("This include file can only be used with the ADwin ProII")
#endif

'<REGION> Constants
#define AB_TRUE 1                 ''<HIDE>
#define AB_FALSE 0                ''<HIDE>
#define AB_UNDEF -1               ''<HIDE>
#define AB_ADR_CR 3FFEh           ''<HIDE>
#define AB_ADR_SR 3FFFh           ''<HIDE>
#define AB_ADR_MOPDW 3800h    'Prozessdaten write area  ''<HIDE>
#define AB_ADR_MOPDR 3900h    'Prozessdaten read area   ''<HIDE>
#define AB_ADR_MOMDW 3B00h    'Message write area       ''<HIDE>
#define AB_ADR_MOMDR 3D00h    'Message read area        ''<HIDE>

#define AB_ADR_MSG_COMM_TYPE 1    ''<HIDE>
#define AB_ADR_MSG_RESP_TYPE 2    ''<HIDE>

'Control register bitmasks.
#define AB_CTRL_T_BIT    80h      ''<HIDE>
#define AB_CTRL_M_BIT    40h      ''<HIDE>
#define AB_CTRL_R_BIT    20h      ''<HIDE>
#define AB_CTRL_AUX_BIT  10h      ''<HIDE>

'Status register bitmasks.
#define AB_STAT_T_BIT    80h      ''<HIDE>
#define AB_STAT_M_BIT    40h      ''<HIDE>
#define AB_STAT_R_BIT    20h      ''<HIDE>
#define AB_STAT_AUX_BIT  10h      ''<HIDE>
#define AB_STAT_SUP_BIT  08h      ''<HIDE>
#define AB_STAT_S_BITS   07h      ''<HIDE>

'Anybus state
#define AB_STAT_SETUP    0        ''<HIDE>
#define AB_STAT_NW_INIT  1        ''<HIDE>
#define AB_STAT_WAIT_PROCESS  2   ''<HIDE>
#define AB_STAT_IDLE     3        ''<HIDE>
#define AB_STAT_PROCESS_ACTIVE 4  ''<HIDE>
#define AB_STAT_ERROR    5        ''<HIDE>
#define AB_STAT_reserved 6        ''<HIDE>
#define AB_STAT_EXCEPTION 7       ''<HIDE>

'Message header bitmasks.
#define AB_MSG_HEAD_C_BIT     40h ''<HIDE>
#define AB_MSG_HEAD_E_BIT     80h ''<HIDE>
#define AB_MSG_HEAD_CMD_BITS  3Fh ''<HIDE>

'HEAD_CMDTYPE
'E bit: 0->Message is a command or successful response
'       1->Message is an Error response
'C bit: 0->Message is a response
'       1->Message is a command

'Message commands.
#define AB_CMD_GET_ATTR          01h    '/* Get attribute                    */ ''<HIDE>
#define AB_CMD_SET_ATTR          02h    '/* Set attribute                    */ ''<HIDE>
#define AB_CMD_CREATE            03h    '/* Create                           */ ''<HIDE>
#define AB_CMD_DELETE            04h    '/* Delete                           */ ''<HIDE>
#define AB_CMD_RESET             05h    '/* Reset                            */ ''<HIDE>
#define AB_CMD_GET_ENUM_STR      06h    '/* Get enumeration string           */ ''<HIDE>
#define AB_CMD_GET_INDEXED_ATTR  07h    '/* Get indexed attribute            */ ''<HIDE>
#define AB_CMD_SET_INDEXED_ATTR  08h    '/* Set indexed attribute            */ ''<HIDE>
#define AB_CMD_MAP_ADI_WRITE_AREA 10h   '/* Map ADI write area               */ ''<HIDE>
#define AB_CMD_MAP_ADI_READ_AREA 11h    '/* Map ADI read area                */ ''<HIDE>

#define AB_BEF_ARR_LEN 200    'Die rx-daten liegen ab idx 1 bis max. 80         ''<HIDE>
#define AB_B_START_IDX 100    'Befehle ab hier                                  ''<HIDE>
#define AB_R_START_IDX 151    'Response ab hier                                 ''<HIDE>

'Arbeitsvariablen
#define AB_L_R_BIT 81     'dim l_r_bit as long                                  ''<HIDE>
#define AB_RX_HEAD_C 82   'dim RX_HEAD_C, RX_HEAD_E as long                     ''<HIDE>
#define AB_RX_HEAD_E 83                                                         ''<HIDE>
#define AB_P_TX 84        'dim p_tx as long                                     ''<HIDE>
#define AB_KOMM_OK 85     'dim komm_ok as long                                  ''<HIDE>
#define AB_ERROR 86       'dim ERROR as long                                    ''<HIDE>
#define AB_F_SEND_RESP 87 'dim f_send_resp, f_send_comm as long                 ''<HIDE>
#define AB_F_SEND_COMM 88                                                       ''<HIDE>
#define AB_LASTSENDMSGTYPE 89 'dim LastSendMsgType as long                      ''<HIDE>
#define AB_STATE_RESPONSE 90  'dim STATE_RESPONSE as long                       ''<HIDE>
#define AB_KOMM_STATE 91      'dim KOMM_STATE as long                           ''<HIDE>
#define AB_L_SR_T 92       'dim l_sr_t, l_sr_t_alt as long                      ''<HIDE>
#define AB_L_SR_T_ALT 93                                                        ''<HIDE>
'<ENDREGION> Constants

'<REGION> System
      
Sub System_Array_Debug_Check(array[], startindex, length)  '<HIDE>
#IF DEBUG_MODE = ON THEN
  if ((startindex) + (length) - 1 > Get_Array_Length(array[1])) then
    PROZESS_FEHLER = 0C1000000H or Get_Array_No(array[1])
    end
  endif
  ''  
  if ((startindex) < 1) then 
    PROZESS_FEHLER = 0C2000000H or Get_Array_No(array[1])
    end
  endif
#endif
ENDSUB

'***************************************************************
''' Converts a process frequency [Hz] into processor ticks (processdelay or cycle time).
''' frequency:  frequency[Hz]
'''*************************************************************
Function Calc_Processdelay(frequency) as long
  
#if prozessor = tico1
  calc_processdelay = 50000000 / (frequency)
#else
  calc_processdelay = 100000000 / (frequency)
#endif
  
endfunction


'***************************************************************
''' Converts timer ticks into ns
''' ticks: counter value ticks into ns
'''*************************************************************
Function Calc_TicksToNs(ticks) As long
  Calc_TicksToNs=(ticks)*20
endfunction





'******************************************************************************
'''Switches the LED on or off.
'''value: 
'''+  0 off
'''+  1 on
'''****************************************************************************
SUB Set_LED(value)
  out (80804000H, value)
ENDSUB


'******************************************************************************
'''Checks the LED.
'''+Return:
'''+   0 off
'''+   1 on
'''****************************************************************************
FUNCTION Check_LED() AS LONG
  Check_LED = in(80804000h)
ENDFUNCTION



'*******************************************************************************
'''Activates or deactivates the external Event input.
'''value:
'''+     0 deactivate
'''+     1 activate
'''*****************************************************************************
SUB Event_Enable(value) ''<HIDE>
  DIM old AS Long

  old = in(80804800h)
  old = (old AND 0FFFFFFFEH) or value
  out (80804800h, old)
ENDSUB

'*******************************************************************************
'''Configures the external Event input.
'''min_hold:
'''+     0 = 15 ns (default) 
'''+     1 = 50 ns
'''edge:
'''+      1 rising edge (default)
'''+      2 falling edge
'''+      3 rising & falling edge
'''prescale: 1 to 255
'''****************************************************************************
SUB Event_Config(min_hold, edge, prescale) ''<HIDE>
  DIM old AS Long
  DIM neu AS Long
  DIM value AS Long

  old = in(80804800h)

  neu = shift_left((min_hold AND 1), 4) or shift_left((edge AND 3), 2)
  old = (old AND 0FFFFFFE3H) or neu
  out (80804800h, old)
  value = prescale AND 0FFH
  out (80804810h, value)
ENDSUB

'*******************************************************************************
'''Creates an event for the ADwin CPU.
'''*****************************************************************************
SUB Trigger_Event()
  out (108041E0h, 1)
ENDSUB


'---------------------------------------------------------------
' If a low-priority process accesses must not be
' interrupted by any other process !!!
'------------------------------------------------------------5700---
Sub Sub_Begin_Critical_Section() ''<HIDE>
#If PRIORITAET = LOW
  Begin_Critical_Section
#EndIf
EndSub

Sub Sub_End_Critical_Section() ''<HIDE>
#If PRIORITAET = LOW
  End_Critical_Section
#EndIf
EndSub

'<ENDREGION>

'<REGION> Anybus

'***************************************************************
'''Gets a byte from the specified memory address of the DPRAM of a profibus-DP master or slave.
'''+return value: content (0..255) of the memory address
'''byteno: number of the byte on the DPRAM
'''***************************************************************
Function Get_Pro_Byte(byteno) As Long ''<HIDE>
  dim in_data as long
  in_data = in (80000000h or byteno)
  Get_Pro_Byte = (in_data and 0FFh)
EndFunction

'***************************************************************
'''Sets a byte in the specified memory address of the DPRAM of a profibus-DP master or slave.
'''+return value: content (0..255) of the memory address
'''byteno: number of the byte on the DPRAM
'''value:  value for the selected byte
'''***************************************************************
Sub Set_Pro_Byte(byteno, value) ''<HIDE>
  out((80000000h or byteno), value and 0FFh)
EndSub

'---------------------------------------------------------------
' Resets hardware on the HMS Module.
' module:  module address
'---------------------------------------------------------------
Sub ABCC_Reset() ''<HIDE>
  
  dim abretry as long
  dim state as long

  abretry = 3

  do
    out(80004000h, 1)'reset active
    sleep(10000) ' =1ms
    out(80004000h, 0)'reset inactive
    sleep(15000000)  '1.5sec (in 10 nsec)
    set_pro_byte(3801H, 15)
    set_pro_byte(3802H, 0)
    
    state = get_pro_byte(3801h)
    dec abretry
  until ((state = 15) or (abretry = 0))
    
  set_pro_byte(3801H, 0)

EndSub

sub write_dat_to_abcc(arr[], arr_idx) ''<HIDE>
  dim i as long
  for i = 0 to arr[arr_idx] - 1
    set_pro_byte(AB_ADR_MOMDW + i, arr[i + arr_idx + 1])   'skip 1. element
  next i
endsub

sub read_msg_header(msg_dat[], start_idx) ''<HIDE>
  dim i as long
  for i = 0 to 7
    msg_dat[i + start_idx] = get_pro_byte(AB_ADR_MOMDR + i)
  next i
endsub

sub read_msg_dat(msg_dat[], md_len, start_idx) ''<HIDE>
  dim i as long
  for i = 0 to md_len - 1
    msg_dat[i + start_idx] = get_pro_byte(AB_ADR_MOMDR + i + 8)
  next i
endsub

sub write_proc_dat(proc_dat[], pd_len) ''<HIDE>
  dim i as long
  for i = 0 to pd_len - 1
    set_pro_byte(AB_ADR_MOPDW + i, proc_dat[i + 1])
  next i
endsub

sub read_proc_dat(proc_dat[], pd_len) ''<HIDE>
  dim i as long
  for i = 0 to pd_len - 1
    proc_dat[i + 1] = get_pro_byte(AB_ADR_MOPDR + i)
  next i
endsub

function get_CTRL() as long ''<HIDE>
  get_CTRL = get_pro_byte(AB_ADR_CR)
endfunction

sub set_CTRL(value) ''<HIDE>
  set_pro_byte(AB_ADR_CR, value)
endsub

function get_STAT() as long ''<HIDE>
  dim t1, t2 as long
  do
    t1 = get_pro_byte(AB_ADR_SR)
    t2 = get_pro_byte(AB_ADR_SR)
  until (t1 = t2)
  get_STAT = t1
endfunction

function get_STAT_T() as long ''<HIDE>
  get_STAT_T = (get_STAT() and AB_STAT_T_BIT)
endfunction

function get_STAT_M() as long ''<HIDE>
  get_STAT_M = (get_STAT() and AB_STAT_M_BIT)
endfunction

function get_STAT_R() as long ''<HIDE>
  get_STAT_R = (get_STAT() and AB_STAT_R_BIT)
endfunction

'<ENDREGION> Anybus

'<REGION> Profibus
'***************************************************************
'''Reads and writes process data from/to Profibus module.
'''+return value: 
'''+         0-> Setup
'''+         2-> Wait
'''+         4-> Active
'''+     other-> Error (configuration correct?)
'''out_pd_arr[]: array with output data
'''out_pd_arr_len: number of output elements
'''in_pd_arr[]: array for input data
'''in_pd_arr_len: number of input elements
'''work_arr[]: array for communication ADwin processor <-> Profibus module 
'''+ the same array as with Init_Profibus
'''***************************************************************
Function Run_Profibus(out_pd_arr[], out_pd_arr_len, in_pd_arr[], in_pd_arr_len, work_arr[]) as long
  dim i, tmp, l_cr as long

  selectcase work_arr[AB_STATE_RESPONSE]
    case 0
      if (work_arr[AB_KOMM_OK] = AB_TRUE) then
        work_arr[AB_STATE_RESPONSE] = 1
      endif
    
    case 1
      if (work_arr[AB_RX_HEAD_C] = AB_TRUE) then
        work_arr[AB_RX_HEAD_C] = AB_UNDEF
        work_arr[AB_L_R_BIT] = AB_FALSE    'R-bit
        work_arr[AB_STATE_RESPONSE] = 2
        work_arr[AB_R_START_IDX] = 8       'len
        work_arr[AB_R_START_IDX + 1] = 19    'Source ID
        work_arr[AB_R_START_IDX + 2] = 1     'objekt
        work_arr[AB_R_START_IDX + 3] = 1     'inst lsb
        work_arr[AB_R_START_IDX + 4] = 0     'inst msb
        work_arr[AB_R_START_IDX + 5] = AB_CMD_GET_ATTR or 80h    'err resp
        work_arr[AB_R_START_IDX + 6] = 0     'Message data size
        work_arr[AB_R_START_IDX + 7] = 1     'Cmd_ext lsb
        work_arr[AB_R_START_IDX + 8] = 0     'Cmd_ext msb
      endif

    case 2
      'Auswerten der Anfrage
      if (work_arr[AB_F_SEND_RESP] = AB_FALSE) then
        work_arr[AB_R_START_IDX] = 8 + work_arr[6]     'resp-len
        for i = 1 to work_arr[AB_R_START_IDX]
          work_arr[AB_R_START_IDX + i] = work_arr[i]
        next i
        work_arr[AB_R_START_IDX + 5] = work_arr[AB_R_START_IDX + 5] and 3Fh  'Err resp
        work_arr[AB_R_START_IDX + 5] = work_arr[AB_R_START_IDX + 5] or 80h   'Err resp
        work_arr[AB_F_SEND_RESP] = AB_TRUE
        work_arr[AB_L_R_BIT] = AB_TRUE
        work_arr[AB_STATE_RESPONSE] = 1
      endif
   
  endselect 
      
  selectcase work_arr[AB_KOMM_STATE]
    case 1

    case 2     'send telegramm with message
      l_cr = get_CTRL()
      l_cr = l_cr or AB_CTRL_M_BIT
      if (work_arr[AB_L_R_BIT] = AB_TRUE) then
        l_cr = l_cr or AB_CTRL_R_BIT          'req to ADWIN
      else
        l_cr = l_cr or (not(AB_CTRL_R_BIT))   'no req to ADWIN 
      endif
      l_cr = l_cr xor AB_CTRL_T_BIT
      write_dat_to_abcc(work_arr, work_arr[AB_P_TX])
      set_CTRL(l_cr)
      work_arr[AB_KOMM_STATE] = 4
    
    case 3     'send telegramm without message
      l_cr = get_CTRL()
      l_cr = l_cr and (not(AB_CTRL_M_BIT))
      if (work_arr[AB_L_R_BIT] = AB_TRUE) then
        l_cr = l_cr or AB_CTRL_R_BIT
      else
        l_cr = l_cr or (not(AB_CTRL_R_BIT))
      endif
      l_cr = l_cr xor AB_CTRL_T_BIT
      write_proc_dat(out_pd_arr, out_pd_arr_len)
      set_CTRL(l_cr)
      work_arr[AB_KOMM_STATE] = 4

    case 4     'wait for response
      work_arr[AB_L_SR_T] = get_stat_t()
      if (work_arr[AB_L_SR_T] <> work_arr[AB_L_SR_T_ALT]) then
        work_arr[AB_L_SR_T_ALT] = work_arr[AB_L_SR_T]
        work_arr[AB_KOMM_STATE] = 5
      endif
    
    case 5    'read response
      tmp = get_stat_m()
      if (tmp = AB_STAT_M_BIT) then
        'new Data
        read_msg_header(work_arr, 1)     'header 8 elements
        read_msg_dat(work_arr, work_arr[6], 9)
        if ((work_arr[5] and AB_MSG_HEAD_C_BIT) = AB_MSG_HEAD_C_BIT) then    'command
          work_arr[AB_RX_HEAD_C] = AB_TRUE
        else                          'response
          work_arr[AB_RX_HEAD_C] = AB_FALSE
        endif
        if ((work_arr[5] and AB_MSG_HEAD_E_BIT) = AB_MSG_HEAD_E_BIT) then    'error resp
          work_arr[AB_RX_HEAD_E] = AB_TRUE
        else                          'command or successful resp
          work_arr[AB_RX_HEAD_E] = AB_FALSE
        endif
      endif
      read_proc_dat(in_pd_arr, in_pd_arr_len)               'read process data
      work_arr[AB_KOMM_STATE] = 6

    case 6   'idle
      work_arr[AB_KOMM_STATE] = 3       'send telegramm without message
      if (get_stat_r() = AB_STAT_R_BIT) then    'ABCC accept more commands
        if (work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_COMM_TYPE) then
          if (work_arr[AB_F_SEND_RESP] = AB_TRUE) then
            'send response
            work_arr[AB_KOMM_STATE] = 2     'send telegramm with message
            work_arr[AB_P_TX] = AB_R_START_IDX     'next response
            work_arr[AB_F_SEND_RESP] = AB_FALSE
            work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_RESP_TYPE
          else
            if (work_arr[AB_F_SEND_COMM] = AB_TRUE) then
              'send command
              work_arr[AB_KOMM_STATE] = 2     'send telegramm with message
              work_arr[AB_P_TX] = AB_B_START_IDX    'next command
              work_arr[AB_F_SEND_COMM] = AB_FALSE
              work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_COMM_TYPE
            endif
          endif
        else
          if (work_arr[AB_F_SEND_COMM] = AB_TRUE) then
            'send command
            work_arr[AB_KOMM_STATE] = 2     'send telegramm with message
            work_arr[AB_P_TX] = AB_B_START_IDX    'next command
            work_arr[AB_F_SEND_COMM] = AB_FALSE
            work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_COMM_TYPE
          else
            if (work_arr[AB_F_SEND_RESP] = AB_TRUE) then
              'send response
              work_arr[AB_KOMM_STATE] = 2     'send telegramm with message
              work_arr[AB_P_TX] = AB_R_START_IDX    'next response
              work_arr[AB_F_SEND_RESP] = AB_FALSE
              work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_RESP_TYPE
            endif 
          endif
        endif
      else                  'ABCC is not ready to accept more commands
        if (work_arr[AB_F_SEND_RESP] = AB_TRUE) then
          'send response
          work_arr[AB_KOMM_STATE] = 2     'send telegramm with message
          work_arr[AB_P_TX] = AB_R_START_IDX    'next response
          work_arr[AB_F_SEND_RESP] = AB_FALSE
          work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_RESP_TYPE
        endif
      endif
      work_arr[AB_KOMM_OK] = AB_TRUE
  endselect
  Run_Profibus = get_stat()

endfunction

'***************************************************************
'''Initializes communication of Profibus module.
'''+return value: 
'''+         0-> OK
'''+     other-> Error
'''value description:
'''  1: error by reading of process data write area
'''  2: error by reading of process data read area
'''  3: error by reading of message write area
'''  4: error by reading of message read area
'''  5: error by mapping of IN data area
'''  6: error by mapping of OUT data area
'''  7: error by setting device address
'''  8: error by reading of module type
'''  9: error by setting setup complete
''' 10: error from Profibus module: Anybus State "ERROR"
''' 11: error from Profibus module: Anybus State "EXCEPTION" 
'''dev_adr: node ID
'''in_mod_cnt: number of input elements
'''in_mod_type: input datatype
'''+         1-> 1 Byte
'''+         2-> 2 Byte
'''+         3-> 4 Byte
'''+        16-> 8 Byte
'''out_mod_cnt: number of output elements
'''out_mod_type: output datatype
'''+         1-> 1 Byte
'''+         2-> 2 Byte
'''+         3-> 4 Byte
'''+        16-> 8 Byte
'''work_arr[]: array for communication ADwin processor <-> Profibus module
'''+  min size 200 elements
'''info[]: info array for bus type
'''+  min size 10 elements
'''+ these return values are correct:
'''+         [1]-> 1
'''+         [2]-> 4
'''***************************************************************
function Init_Profibus(dev_adr, in_mod_cnt, in_mod_type, out_mod_cnt, out_mod_type, work_arr[], info[]) as long
  dim init_fertig as long
  dim ret as long
  dim STATE_BEFEHLE as long
  dim i, tmp as long
  
  dim l_in_pd_arr[76] as long
  dim l_out_pd_arr[76] as long
  
  ABCC_Reset()

  for i = 1 to AB_BEF_ARR_LEN
    work_arr[i] = 0
  next i

  ret = AB_FALSE
  init_fertig = AB_FALSE
  work_arr[AB_ERROR] = 0
  'check Profibusmodul
  for i = 0 to 0FFh
    set_pro_byte(AB_ADR_MOPDW + i, i + 1)    'process data write area
    set_pro_byte(AB_ADR_MOPDR + i, i + 2)    'process data read area
    set_pro_byte(AB_ADR_MOMDW + i, i + 3)    'Message write area
    set_pro_byte(AB_ADR_MOMDR + i, i + 4)    'Message read area
  next i
  for i = 0 to 0FFh
    tmp = get_pro_byte(AB_ADR_MOPDW + i)
    if (tmp <> (i + 1) and 0FFh) then work_arr[AB_ERROR] = 1
    tmp = get_pro_byte(AB_ADR_MOPDR + i)
    if (tmp <> (i + 2) and 0FFh) then work_arr[AB_ERROR] = 2
    tmp = get_pro_byte(AB_ADR_MOMDW + i)
    if (tmp <> (i + 3) and 0FFh) then work_arr[AB_ERROR] = 3
    tmp = get_pro_byte(AB_ADR_MOMDR + i)
    if (tmp <> (i + 4) and 0FFh) then work_arr[AB_ERROR] = 4
  next i
  work_arr[AB_KOMM_STATE] = 3
  work_arr[AB_L_SR_T] = get_STAT_T()
  work_arr[AB_L_SR_T_ALT] = work_arr[AB_L_SR_T]
  STATE_BEFEHLE = 0
  work_arr[AB_P_TX] = AB_B_START_IDX
  work_arr[AB_STATE_RESPONSE] = 0
  work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_RESP_TYPE
  work_arr[AB_F_SEND_RESP] = AB_FALSE
  work_arr[AB_F_SEND_COMM] = AB_FALSE
  work_arr[AB_RX_HEAD_C] = AB_UNDEF
  work_arr[AB_RX_HEAD_E] = AB_UNDEF
  work_arr[AB_L_R_BIT] = AB_TRUE
  work_arr[AB_KOMM_OK] = AB_FALSE

  do
    selectcase STATE_BEFEHLE
      case 0
        if (work_arr[AB_KOMM_OK] = AB_TRUE) then
          STATE_BEFEHLE = 1
        endif
    
      case 1
        'data_type
        '0: bool 8bit
        '1: sint 8bit
        '2: sint 16bit
        '3: sint 32bit
        '4: uint 8bit
        '5: uint 16bit
        '6: uint 32
        '7: uint char 8bit
        '8: enum 8bit
        '16: sint 64bit
        '17: uint 64nit
        '18 float 32bit
        if (work_arr[AB_F_SEND_COMM] = AB_FALSE) then
          'B_IDX_MAP_ADI_READ_AREA
          work_arr[AB_B_START_IDX] = 12        'len
          work_arr[AB_B_START_IDX + 1] = 1     'Source ID
          work_arr[AB_B_START_IDX + 2] = 3     'objekt
          work_arr[AB_B_START_IDX + 3] = 1     'inst lsb
          work_arr[AB_B_START_IDX + 4] = 0     'inst msb
          work_arr[AB_B_START_IDX + 5] = AB_CMD_MAP_ADI_READ_AREA or 40h    'command
          work_arr[AB_B_START_IDX + 6] = 4     'Message data size
          work_arr[AB_B_START_IDX + 7] = 3     'Cmd_ext lsb
          work_arr[AB_B_START_IDX + 8] = 0     'Cmd_ext msb
          work_arr[AB_B_START_IDX + 9] = IN_MOD_TYPE  '1.data  data_type
          work_arr[AB_B_START_IDX + 10] = IN_MOD_CNT  '2.number of elements
          work_arr[AB_B_START_IDX + 11] = 1    '3.      order no low byte
          work_arr[AB_B_START_IDX + 12] = 0    '4.      order no high byte
          work_arr[AB_F_SEND_COMM] = AB_TRUE
          STATE_BEFEHLE = 2
        endif
    
      case 2
        if (work_arr[AB_RX_HEAD_C] = AB_FALSE) then
          work_arr[AB_RX_HEAD_C] = AB_UNDEF
          if (work_arr[AB_RX_HEAD_E] = AB_FALSE) then
            STATE_BEFEHLE = 3
          else
            work_arr[AB_ERROR] = 5
          endif
        endif

      case 3
        if (work_arr[AB_F_SEND_COMM] = AB_FALSE) then
          'B_IDX_MAP_ADI_WRITE_AREA
          work_arr[AB_B_START_IDX] = 12
          work_arr[AB_B_START_IDX + 1] = 3
          work_arr[AB_B_START_IDX + 2] = 3
          work_arr[AB_B_START_IDX + 3] = 1
          work_arr[AB_B_START_IDX + 4] = 0
          work_arr[AB_B_START_IDX + 5] = AB_CMD_MAP_ADI_WRITE_AREA or 40h
          work_arr[AB_B_START_IDX + 6] = 4
          work_arr[AB_B_START_IDX + 7] = 3
          work_arr[AB_B_START_IDX + 8] = 0
          work_arr[AB_B_START_IDX + 9] = OUT_MOD_TYPE
          work_arr[AB_B_START_IDX + 10] = OUT_MOD_CNT
          work_arr[AB_B_START_IDX + 11] = 1
          work_arr[AB_B_START_IDX + 12] = 0
          work_arr[AB_F_SEND_COMM] = AB_TRUE
          STATE_BEFEHLE = 4
        endif

      case 4
        if (work_arr[AB_RX_HEAD_C] = AB_FALSE) then
          work_arr[AB_RX_HEAD_C] = AB_UNDEF
          if (work_arr[AB_RX_HEAD_E] = AB_FALSE) then
            STATE_BEFEHLE = 5
          else
            work_arr[AB_ERROR] = 6
          endif
        endif

      case 5
        if (work_arr[AB_F_SEND_COMM] = AB_FALSE) then
          'B_SET_DEV_ADR
          work_arr[AB_B_START_IDX] = 9
          work_arr[AB_B_START_IDX + 1] = 5
          work_arr[AB_B_START_IDX + 2] = 4
          work_arr[AB_B_START_IDX + 3] = 1
          work_arr[AB_B_START_IDX + 4] = 0
          work_arr[AB_B_START_IDX + 5] = AB_CMD_SET_ATTR or 40h
          work_arr[AB_B_START_IDX + 6] = 1
          work_arr[AB_B_START_IDX + 7] = 5
          work_arr[AB_B_START_IDX + 8] = 0
          work_arr[AB_B_START_IDX + 9] = DEV_ADR
          work_arr[AB_F_SEND_COMM] = AB_TRUE
          STATE_BEFEHLE = 6
        endif

      case 6
        if (work_arr[AB_RX_HEAD_C] = AB_FALSE) then
          work_arr[AB_RX_HEAD_C] = AB_UNDEF
          if (work_arr[AB_RX_HEAD_E] = AB_FALSE) then
            STATE_BEFEHLE = 7
          else
            work_arr[AB_ERROR] = 7
          endif
        endif

      case 7
        if (work_arr[AB_F_SEND_COMM] = AB_FALSE) then
          'REQ_READ_MODULE_TYPE
          work_arr[AB_B_START_IDX] = 8
          work_arr[AB_B_START_IDX + 1] = 11
          work_arr[AB_B_START_IDX + 2] = 1
          work_arr[AB_B_START_IDX + 3] = 1
          work_arr[AB_B_START_IDX + 4] = 0
          work_arr[AB_B_START_IDX + 5] = AB_CMD_GET_ATTR or 40h
          work_arr[AB_B_START_IDX + 6] = 0
          work_arr[AB_B_START_IDX + 7] = 1
          work_arr[AB_B_START_IDX + 8] = 0
          work_arr[AB_F_SEND_COMM] = AB_TRUE
          STATE_BEFEHLE = 8
        endif

      case 8
        if (work_arr[AB_RX_HEAD_C] = AB_FALSE) then
          work_arr[AB_RX_HEAD_C] = AB_UNDEF
          if (work_arr[AB_RX_HEAD_E] = AB_FALSE) then
            for i = 1 to 10
              info[i] = work_arr[i + 8]
            next i
            STATE_BEFEHLE = 9
          else
            work_arr[AB_ERROR] = 8
          endif
        endif

      case 9
        if (work_arr[AB_F_SEND_COMM] = AB_FALSE) then
          'B_SETUP_COMPLETE
          work_arr[AB_B_START_IDX] = 9
          work_arr[AB_B_START_IDX + 1] = 9
          work_arr[AB_B_START_IDX + 2] = 1
          work_arr[AB_B_START_IDX + 3] = 1
          work_arr[AB_B_START_IDX + 4] = 0
          work_arr[AB_B_START_IDX + 5] = AB_CMD_SET_ATTR or 40h
          work_arr[AB_B_START_IDX + 6] = 1
          work_arr[AB_B_START_IDX + 7] = 5
          work_arr[AB_B_START_IDX + 8] = 0
          work_arr[AB_B_START_IDX + 9] = 1
          work_arr[AB_F_SEND_COMM] = AB_TRUE
          STATE_BEFEHLE = 10
        endif

      case 10
        if (work_arr[AB_RX_HEAD_C] = AB_FALSE) then
          work_arr[AB_RX_HEAD_C] = AB_UNDEF
          if (work_arr[AB_RX_HEAD_E] = AB_FALSE) then
            STATE_BEFEHLE = 11
          else
            work_arr[AB_ERROR] = 9
          endif
        endif

      case 11
    
    endselect
    tmp = Run_Profibus(l_out_pd_arr, 76, l_in_pd_arr, 76, work_arr) and 7h
    if (tmp = 2) then init_fertig = AB_TRUE
    if (tmp = 5) then work_arr[AB_ERROR] = 10               'error
    if (tmp = 7) then work_arr[AB_ERROR] = 11               'exception

  until ((init_fertig = AB_TRUE) or (work_arr[AB_ERROR] <> 0))
  Init_Profibus = work_arr[AB_ERROR]
endfunction
'<ENDREGION> Profibus

'<REGION> DeviceNet
'***************************************************************
'''Reads and writes process data from/to DeviceNet module.
'''+return value: 
'''+         0-> Setup
'''+         2-> Wait
'''+         4-> Active
'''+     other-> Error (configuration correct?)
'''in_pd_arr[]: array for input data
'''in_pd_arr_len: number of input elements
'''out_pd_arr[]: array with output data
'''out_pd_arr_len: number of output elements
'''work_arr[]: array for communication ADwin processor <-> DeviceNet module 
'''+ the same as at Init function
'''***************************************************************
function Run_DeviceNet(in_pd_arr[], in_pd_arr_len, out_pd_arr[], out_pd_arr_len, work_arr[]) as long
  dim i, tmp, l_cr as long

  selectcase work_arr[AB_STATE_RESPONSE]
    case 0
      if (work_arr[AB_KOMM_OK] = AB_TRUE) then
        work_arr[AB_STATE_RESPONSE] = 1
      endif
    
    case 1
      if (work_arr[AB_RX_HEAD_C] = AB_TRUE) then
        work_arr[AB_RX_HEAD_C] = AB_UNDEF
        work_arr[AB_L_R_BIT] = AB_FALSE    'R-bit
        work_arr[AB_STATE_RESPONSE] = 2
        work_arr[AB_R_START_IDX] = 8       'len
        work_arr[AB_R_START_IDX + 1] = 19    'Source ID
        work_arr[AB_R_START_IDX + 2] = 1     'objekt
        work_arr[AB_R_START_IDX + 3] = 1     'inst lsb
        work_arr[AB_R_START_IDX + 4] = 0     'inst msb
        work_arr[AB_R_START_IDX + 5] = AB_CMD_GET_ATTR or 80h    'err resp
        work_arr[AB_R_START_IDX + 6] = 0     'Message data size
        work_arr[AB_R_START_IDX + 7] = 1     'Cmd_ext lsb
        work_arr[AB_R_START_IDX + 8] = 0     'Cmd_ext msb
      endif

    case 2
      'Auswerten der Anfrage
      if (work_arr[AB_F_SEND_RESP] = AB_FALSE) then
        work_arr[AB_R_START_IDX] = 8 + work_arr[6]     'resp-len
        for i = 1 to work_arr[AB_R_START_IDX]
          work_arr[AB_R_START_IDX + i] = work_arr[i]
        next i
        work_arr[AB_R_START_IDX + 5] = work_arr[AB_R_START_IDX + 5] and 3Fh  'Err resp
        work_arr[AB_R_START_IDX + 5] = work_arr[AB_R_START_IDX + 5] or 80h   'Err resp
        work_arr[AB_F_SEND_RESP] = AB_TRUE
        work_arr[AB_L_R_BIT] = AB_TRUE
        work_arr[AB_STATE_RESPONSE] = 1
      endif
   
  endselect 
      
  selectcase work_arr[AB_KOMM_STATE]
    case 1

    case 2     'send telegramm with message
      l_cr = get_CTRL()
      l_cr = l_cr or AB_CTRL_M_BIT
      if (work_arr[AB_L_R_BIT] = AB_TRUE) then
        l_cr = l_cr or AB_CTRL_R_BIT          'req to ADWIN
      else
        l_cr = l_cr or (not(AB_CTRL_R_BIT))   'no req to ADWIN 
      endif
      l_cr = l_cr xor AB_CTRL_T_BIT
      write_dat_to_abcc(work_arr, work_arr[AB_P_TX])
      set_CTRL(l_cr)
      work_arr[AB_KOMM_STATE] = 4
    
    case 3     'send telegramm without message
      l_cr = get_CTRL()
      l_cr = l_cr and (not(AB_CTRL_M_BIT))
      if (work_arr[AB_L_R_BIT] = AB_TRUE) then
        l_cr = l_cr or AB_CTRL_R_BIT
      else
        l_cr = l_cr or (not(AB_CTRL_R_BIT))
      endif
      l_cr = l_cr xor AB_CTRL_T_BIT
      write_proc_dat(out_pd_arr, out_pd_arr_len)
      set_CTRL(l_cr)
      work_arr[AB_KOMM_STATE] = 4

    case 4     'wait for response
      work_arr[AB_L_SR_T] = get_stat_t()
      if (work_arr[AB_L_SR_T] <> work_arr[AB_L_SR_T_ALT]) then
        work_arr[AB_L_SR_T_ALT] = work_arr[AB_L_SR_T]
        work_arr[AB_KOMM_STATE] = 5
      endif
    
    case 5    'read response
      tmp = get_stat_m()
      if (tmp = AB_STAT_M_BIT) then
        'new Data
        read_msg_header(work_arr, 1)     'header 8 elements
        read_msg_dat(work_arr, work_arr[6], 9)
        if ((work_arr[5] and AB_MSG_HEAD_C_BIT) = AB_MSG_HEAD_C_BIT) then    'command
          work_arr[AB_RX_HEAD_C] = AB_TRUE
        else                          'response
          work_arr[AB_RX_HEAD_C] = AB_FALSE
        endif
        if ((work_arr[5] and AB_MSG_HEAD_E_BIT) = AB_MSG_HEAD_E_BIT) then    'error resp
          work_arr[AB_RX_HEAD_E] = AB_TRUE
        else                          'command or successful resp
          work_arr[AB_RX_HEAD_E] = AB_FALSE
        endif
      endif
      read_proc_dat(in_pd_arr, in_pd_arr_len)               'read process data
      work_arr[AB_KOMM_STATE] = 6

    case 6   'idle
      work_arr[AB_KOMM_STATE] = 3       'send telegramm without message
      if (get_stat_r() = AB_STAT_R_BIT) then    'ABCC accept more commands
        if (work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_COMM_TYPE) then
          if (work_arr[AB_F_SEND_RESP] = AB_TRUE) then
            'send response
            work_arr[AB_KOMM_STATE] = 2     'send telegramm with message
            work_arr[AB_P_TX] = AB_R_START_IDX     'next response
            work_arr[AB_F_SEND_RESP] = AB_FALSE
            work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_RESP_TYPE
          else
            if (work_arr[AB_F_SEND_COMM] = AB_TRUE) then
              'send command
              work_arr[AB_KOMM_STATE] = 2     'send telegramm with message
              work_arr[AB_P_TX] = AB_B_START_IDX    'next command
              work_arr[AB_F_SEND_COMM] = AB_FALSE
              work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_COMM_TYPE
            endif
          endif
        else
          if (work_arr[AB_F_SEND_COMM] = AB_TRUE) then
            'send command
            work_arr[AB_KOMM_STATE] = 2     'send telegramm with message
            work_arr[AB_P_TX] = AB_B_START_IDX    'next command
            work_arr[AB_F_SEND_COMM] = AB_FALSE
            work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_COMM_TYPE
          else
            if (work_arr[AB_F_SEND_RESP] = AB_TRUE) then
              'send response
              work_arr[AB_KOMM_STATE] = 2     'send telegramm with message
              work_arr[AB_P_TX] = AB_R_START_IDX    'next response
              work_arr[AB_F_SEND_RESP] = AB_FALSE
              work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_RESP_TYPE
            endif 
          endif
        endif
      else                  'ABCC is not ready to accept more commands
        if (work_arr[AB_F_SEND_RESP] = AB_TRUE) then
          'send response
          work_arr[AB_KOMM_STATE] = 2     'send telegramm with message
          work_arr[AB_P_TX] = AB_R_START_IDX    'next response
          work_arr[AB_F_SEND_RESP] = AB_FALSE
          work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_RESP_TYPE
        endif
      endif
      work_arr[AB_KOMM_OK] = AB_TRUE
  endselect
  Run_DeviceNet = get_stat()

endfunction

'***************************************************************
'''Initializes communication of DeviceNet module.
'''+return value: 
'''+         0-> OK
'''+     other-> Error
'''value description:
'''  1: error by reading of process data write area
'''  2: error by reading of process data read area
'''  3: error by reading of message write area
'''  4: error by reading of message read area
'''  5: error by mapping of IN data area
'''  6: error by mapping of OUT data area
'''  7: error by setting node ID
'''  8: error by setting baud rate
'''  9: error by reading of module type
''' 10: error by setting setup complete
''' 11: error from DeviceNet module: Anybus State "ERROR"
''' 12: error from DeviceNet module: Anybus State "EXCEPTION" 
'''dev_adr: node ID
'''baudrate: 
'''+         0-> 125kbit
'''+         1-> 250kbit
'''+         2-> 250kbit
'''+         3-> Autobaud
'''in_mod_cnt: number of input elements
'''in_mod_type: input datatype
'''+         1-> 1 Byte
'''+         2-> 2 Byte
'''+         3-> 4 Byte
'''+        16-> 8 Byte
'''out_mod_cnt: number of output elements
'''out_mod_type: output datatype
'''+         1-> 1 Byte
'''+         2-> 2 Byte
'''+         3-> 4 Byte
'''+        16-> 8 Byte
'''work_arr[]: array for communication ADwin processor <-> DeviceNet module
'''+  min size 200 elements
'''info[]: info array for bustype
'''+  min size 10 elements
'''+ these returnvalues are correct:
'''+         [1]-> 1
'''+         [2]-> 4
'''***************************************************************
function Init_DeviceNet(dev_adr, baudrate, in_mod_cnt, in_mod_type, out_mod_cnt, out_mod_type, work_arr[], info[]) as long
  dim init_fertig as long
  dim ret as long
  dim STATE_BEFEHLE as long
  dim i, tmp as long
  
  dim l_in_pd_arr[256] as long
  dim l_out_pd_arr[256] as long
  
  ABCC_Reset()

  for i = 1 to AB_BEF_ARR_LEN
    work_arr[i] = 0
  next i

  ret = AB_FALSE
  init_fertig = AB_FALSE
  work_arr[AB_ERROR] = 0
  'check DeviceNetmodul
  for i = 0 to 0FFh
    set_pro_byte(AB_ADR_MOPDW + i, i + 1)    'process data write area
    set_pro_byte(AB_ADR_MOPDR + i, i + 2)    'process data read area
    set_pro_byte(AB_ADR_MOMDW + i, i + 3)    'Message write area
    set_pro_byte(AB_ADR_MOMDR + i, i + 4)    'Message read area
  next i
  for i = 0 to 0FFh
    tmp = get_pro_byte(AB_ADR_MOPDW + i)
    if (tmp <> (i + 1) and 0FFh) then work_arr[AB_ERROR] = 1
    tmp = get_pro_byte(AB_ADR_MOPDR + i)
    if (tmp <> (i + 2) and 0FFh) then work_arr[AB_ERROR] = 2
    tmp = get_pro_byte(AB_ADR_MOMDW + i)
    if (tmp <> (i + 3) and 0FFh) then work_arr[AB_ERROR] = 3
    tmp = get_pro_byte(AB_ADR_MOMDR + i)
    if (tmp <> (i + 4) and 0FFh) then work_arr[AB_ERROR] = 4
  next i
  work_arr[AB_KOMM_STATE] = 3
  work_arr[AB_L_SR_T] = get_STAT_T()
  work_arr[AB_L_SR_T_ALT] = work_arr[AB_L_SR_T]
  STATE_BEFEHLE = 0
  work_arr[AB_P_TX] = AB_B_START_IDX
  work_arr[AB_STATE_RESPONSE] = 0
  work_arr[AB_LASTSENDMSGTYPE] = AB_ADR_MSG_RESP_TYPE
  work_arr[AB_F_SEND_RESP] = AB_FALSE
  work_arr[AB_F_SEND_COMM] = AB_FALSE
  work_arr[AB_RX_HEAD_C] = AB_UNDEF
  work_arr[AB_RX_HEAD_E] = AB_UNDEF
  work_arr[AB_L_R_BIT] = AB_TRUE
  work_arr[AB_KOMM_OK] = AB_FALSE

  do
    selectcase STATE_BEFEHLE
      case 0
        if (work_arr[AB_KOMM_OK] = AB_TRUE) then
          STATE_BEFEHLE = 1
        endif
    
      case 1
        'data_type
        '0: bool 8bit
        '1: sint 8bit
        '2: sint 16bit
        '3: sint 32bit
        '4: uint 8bit
        '5: uint 16bit
        '6: uint 32
        '7: uint char 8bit
        '8: enum 8bit
        '16: sint 64bit
        '17: uint 64nit
        '18 float 32bit
        if (work_arr[AB_F_SEND_COMM] = AB_FALSE) then
          'B_IDX_MAP_ADI_READ_AREA
          work_arr[AB_B_START_IDX] = 12        'len
          work_arr[AB_B_START_IDX + 1] = 1     'Source ID
          work_arr[AB_B_START_IDX + 2] = 3     'objekt
          work_arr[AB_B_START_IDX + 3] = 1     'inst lsb
          work_arr[AB_B_START_IDX + 4] = 0     'inst msb
          work_arr[AB_B_START_IDX + 5] = AB_CMD_MAP_ADI_READ_AREA or 40h    'command
          work_arr[AB_B_START_IDX + 6] = 4     'Message data size
          work_arr[AB_B_START_IDX + 7] = 3     'Cmd_ext lsb
          work_arr[AB_B_START_IDX + 8] = 0     'Cmd_ext msb
          work_arr[AB_B_START_IDX + 9] = IN_MOD_TYPE  '1.data  data_type
          work_arr[AB_B_START_IDX + 10] = IN_MOD_CNT  '2.number of elements
          work_arr[AB_B_START_IDX + 11] = 1    '3.      order no low byte
          work_arr[AB_B_START_IDX + 12] = 0    '4.      order no high byte
          work_arr[AB_F_SEND_COMM] = AB_TRUE
          STATE_BEFEHLE = 2
        endif
    
      case 2
        if (work_arr[AB_RX_HEAD_C] = AB_FALSE) then
          work_arr[AB_RX_HEAD_C] = AB_UNDEF
          if (work_arr[AB_RX_HEAD_E] = AB_FALSE) then
            STATE_BEFEHLE = 3
          else
            work_arr[AB_ERROR] = 5
          endif
        endif

      case 3
        if (work_arr[AB_F_SEND_COMM] = AB_FALSE) then
          'B_IDX_MAP_ADI_WRITE_AREA
          work_arr[AB_B_START_IDX] = 12
          work_arr[AB_B_START_IDX + 1] = 3
          work_arr[AB_B_START_IDX + 2] = 3
          work_arr[AB_B_START_IDX + 3] = 1
          work_arr[AB_B_START_IDX + 4] = 0
          work_arr[AB_B_START_IDX + 5] = AB_CMD_MAP_ADI_WRITE_AREA or 40h
          work_arr[AB_B_START_IDX + 6] = 4
          work_arr[AB_B_START_IDX + 7] = 3
          work_arr[AB_B_START_IDX + 8] = 0
          work_arr[AB_B_START_IDX + 9] = OUT_MOD_TYPE
          work_arr[AB_B_START_IDX + 10] = OUT_MOD_CNT
          work_arr[AB_B_START_IDX + 11] = 1
          work_arr[AB_B_START_IDX + 12] = 0
          work_arr[AB_F_SEND_COMM] = AB_TRUE
          STATE_BEFEHLE = 4
        endif

      case 4
        if (work_arr[AB_RX_HEAD_C] = AB_FALSE) then
          work_arr[AB_RX_HEAD_C] = AB_UNDEF
          if (work_arr[AB_RX_HEAD_E] = AB_FALSE) then
            STATE_BEFEHLE = 5
          else
            work_arr[AB_ERROR] = 6
          endif
        endif

      case 5
        if (work_arr[AB_F_SEND_COMM] = AB_FALSE) then
          ' Node ID (MacID)
          work_arr[AB_B_START_IDX] = 9
          work_arr[AB_B_START_IDX + 1] = 5
          work_arr[AB_B_START_IDX + 2] = 4
          work_arr[AB_B_START_IDX + 3] = 1
          work_arr[AB_B_START_IDX + 4] = 0
          work_arr[AB_B_START_IDX + 5] = AB_CMD_SET_ATTR or 40h
          work_arr[AB_B_START_IDX + 6] = 1
          work_arr[AB_B_START_IDX + 7] = 5
          work_arr[AB_B_START_IDX + 8] = 0
          work_arr[AB_B_START_IDX + 9] = DEV_ADR
          work_arr[AB_F_SEND_COMM] = AB_TRUE
          STATE_BEFEHLE = 6
        endif

      case 6
        if (work_arr[AB_RX_HEAD_C] = AB_FALSE) then
          work_arr[AB_RX_HEAD_C] = AB_UNDEF
          if (work_arr[AB_RX_HEAD_E] = AB_FALSE) then
            STATE_BEFEHLE = 7
          else
            work_arr[AB_ERROR] = 7
          endif
        endif

      case 7
        if (work_arr[AB_F_SEND_COMM] = AB_FALSE) then
          ' Baud_Rate
          ' 0-> 125kbit
          ' 1-> 250kbit
          ' 2-> 500kbit
          ' 3-> Autobaud
          work_arr[AB_B_START_IDX] = 9
          work_arr[AB_B_START_IDX + 1] = 4  'Source ID
          work_arr[AB_B_START_IDX + 2] = 4  'objekt
          work_arr[AB_B_START_IDX + 3] = 2  'inst lsb
          work_arr[AB_B_START_IDX + 4] = 0  'inst msb
          work_arr[AB_B_START_IDX + 5] = AB_CMD_SET_ATTR or 40h  'command
          work_arr[AB_B_START_IDX + 6] = 1  'Message data size
          work_arr[AB_B_START_IDX + 7] = 5  'Cmd_ext lsb
          work_arr[AB_B_START_IDX + 8] = 0  'Cmd_ext msb
          work_arr[AB_B_START_IDX + 9] = baudrate  '1.data
          work_arr[AB_F_SEND_COMM] = AB_TRUE
          STATE_BEFEHLE = 8
        endif

      case 8
        if (work_arr[AB_RX_HEAD_C] = AB_FALSE) then
          work_arr[AB_RX_HEAD_C] = AB_UNDEF
          if (work_arr[AB_RX_HEAD_E] = AB_FALSE) then
            STATE_BEFEHLE = 9
          else
            work_arr[AB_ERROR] = 8
          endif
        endif

      case 9
        if (work_arr[AB_F_SEND_COMM] = AB_FALSE) then
          'REQ_READ_MODULE_TYPE
          work_arr[AB_B_START_IDX] = 8
          work_arr[AB_B_START_IDX + 1] = 11
          work_arr[AB_B_START_IDX + 2] = 1
          work_arr[AB_B_START_IDX + 3] = 1
          work_arr[AB_B_START_IDX + 4] = 0
          work_arr[AB_B_START_IDX + 5] = AB_CMD_GET_ATTR or 40h
          work_arr[AB_B_START_IDX + 6] = 0
          work_arr[AB_B_START_IDX + 7] = 1
          work_arr[AB_B_START_IDX + 8] = 0
          work_arr[AB_F_SEND_COMM] = AB_TRUE
          STATE_BEFEHLE = 10
        endif

      case 10
        if (work_arr[AB_RX_HEAD_C] = AB_FALSE) then
          work_arr[AB_RX_HEAD_C] = AB_UNDEF
          if (work_arr[AB_RX_HEAD_E] = AB_FALSE) then
            for i = 1 to 10
              info[i] = work_arr[i + 8]
            next i
            STATE_BEFEHLE = 11
          else
            work_arr[AB_ERROR] = 9
          endif
        endif

      case 11
        if (work_arr[AB_F_SEND_COMM] = AB_FALSE) then
          'B_SETUP_COMPLETE
          work_arr[AB_B_START_IDX] = 9
          work_arr[AB_B_START_IDX + 1] = 9
          work_arr[AB_B_START_IDX + 2] = 1
          work_arr[AB_B_START_IDX + 3] = 1
          work_arr[AB_B_START_IDX + 4] = 0
          work_arr[AB_B_START_IDX + 5] = AB_CMD_SET_ATTR or 40h
          work_arr[AB_B_START_IDX + 6] = 1
          work_arr[AB_B_START_IDX + 7] = 5
          work_arr[AB_B_START_IDX + 8] = 0
          work_arr[AB_B_START_IDX + 9] = 1
          work_arr[AB_F_SEND_COMM] = AB_TRUE
          STATE_BEFEHLE = 12
        endif

      case 12
        if (work_arr[AB_RX_HEAD_C] = AB_FALSE) then
          work_arr[AB_RX_HEAD_C] = AB_UNDEF
          if (work_arr[AB_RX_HEAD_E] = AB_FALSE) then
            STATE_BEFEHLE = 13
          else
            work_arr[AB_ERROR] = 10
          endif
        endif

      case 13
    
    endselect
    tmp = Run_DeviceNet(l_in_pd_arr, 256, l_out_pd_arr, 256, work_arr) and 7h
    if (tmp = 2) then init_fertig = AB_TRUE
    if (tmp = 5) then work_arr[AB_ERROR] = 11               'error
    if (tmp = 7) then work_arr[AB_ERROR] = 12               'exception

  until ((init_fertig = AB_TRUE) or (work_arr[AB_ERROR] <> 0))
  Init_DeviceNet = work_arr[AB_ERROR]
endfunction

'<ENDREGION> DeviceNet

'<REGION> Anybus M40

SUB ABM40_Reset() '<HIDE>
  out(4000H, 1) 'Reset low '1  ms
  sleep(10000)  
  out(4000H, 0) 'Reset high
  sleep(15000000)                                     '1.5sec (in 10 nsec)
ENDSUB
    
'''read RAM content
'''destination_ar type long
'''startindex = startindex destination array 1....
'''srartindex_ram 0.... (word address)
'''count 1.. 
'''+  word_length = 16 => count = number of 16 bit words 
'''+  word_length = 32 => count= number of 32 bit words
'''flowrate: 
'''+   low priority 1 =slow (up to 16 x long) , 2 middle (up to 128 x long) , 3=fast ( up to 1024 x long)
'''+   high priority 1,2,3 (up to 32768 x long)
'''word_length:
'''   32   value use 32 bit in destination array 
'''   16   value use 16 bit in destination array 
SUB M40_GET_RAM(destination_ar[], startindex, startindex_ram, count, flowrate, word_length) '<HIDE>
  dim anzcount as integer
  dim icount as long
  DIM iindex as integer
  
  System_Array_Debug_Check(destination_ar, startindex, count)
#IF PRIORITAET = LOW
  DIM maxburst as Long
  DIM i as long
#endif    
  
  if (word_length = 32) then 
    out(7085h, (startindex_ram) or shift_left(1, 31))
#IF PRIORITAET = HIGH
    nops(6)
#else
    nops(4)    
#endif

  else
    out(7085h, startindex_ram) 
#IF PRIORITAET = HIGH
    nops(4)
#else
    nops(2)
#endif


  endif


  
  if (count > 0) then 
    iindex = startindex
   
    
#IF PRIORITAET = LOW
    maxburst = 128
    if (flowrate = 1) then maxburst = 16
    if (flowrate = 3) then maxburst = 1024
    icount = count
   
    do
      anzcount = min_long(icount, maxburst) 
      
      for i = 1 to anzcount
        destination_ar[iindex] = in(5000H)
        inc iindex
      next
      
      icount = icount - anzcount
   
    until (icount = 0)
      
      
#else
                 
    anzcount = count 
      
    if (word_length = 32) then 
      
      begin_critical_section ()     
      asm{
      
      ldc destination_ar ''loads the start pointer of Data in the A Register      
      ldi @___M40_GET_RAM@iindex'loads the startindex in A Register (start pointer of Data_1 will be shifted to B register)
      add   'adds A and B register      
      stindex2 (nop)   'Index register2 points to data_1[startindex], to start from this point      
     
      ldi @___M40_GET_RAM@anzcount 'loads numvalues in A Register
     
       
      loop ma_loop_WDL1G + #Macro_call_count  'loops with "numvalues" repitions
      '------------ start of the loop
      
      in 5000H       'input to A Register
      sti2+ (nop)   'value from A Register to array and Increment index Register
             
      nop 
      nop
        
      '-----------  'End of the loop 
      ma_loop_WDL1G + #Macro_call_count :
      }end_asm 
      end_critical_section ()

    else
      begin_critical_section ()     
      asm{
      
      ldc destination_ar ''loads the start pointer of Data in the A Register      
      ldi @___M40_GET_RAM@iindex'loads the startindex in A Register (start pointer of Data_1 will be shifted to B register)
      add   'adds A and B register      
      stindex2 (nop)   'Index register2 points to data_1[startindex], to start from this point      
     
      ldi @___M40_GET_RAM@anzcount 'loads numvalues in A Register
     
       
      loop ma_loop_WDL1G16 + #Macro_call_count  'loops with "numvalues" repitions
      '------------ start of the loop
      
      in 5000H       'input to A Register
      sti2+ (nop)   'value from A Register to array and Increment index Register
                
      '-----------  'End of the loop 
      ma_loop_WDL1G16 + #Macro_call_count :
      }end_asm 
      end_critical_section ()
        
    endif
    
#endif       
        
 
  endif
  
  
  
ENDSUB

'''set RAM content
'''source_ar: source array type long
'''startindex = startindex source array 1....
'''srartindex ram 0....
'''count 1....
'''+  word_length = 16 => count = number of 16 bit words 
'''+  word_length = 32 => count= number of 32 bit words
'''flowrate 
'''+   low priority 1 =slow (up to 16 x long) , 2 middle (up to 128 x long) , 3=fast ( up to 1024 x long)
'''+   high priority 1,2,3 (up to 32768 x long)
'''word_length:
'''   32   value use 32 bit in source array 
'''   16   value use 16 bit in source array 
SUB M40_SET_RAM(source_ar[], startindex, startindex_ram, count, flowrate, word_length) '<HIDE>
  dim anzcount as Long
  dim icount as long
  DIM iindex as Long
  dim ready as long
  
  
  System_Array_Debug_Check(source_ar, startindex, count)
#if PRIORITAET = LOW
  DIM maxburst as Long
  DIM I as long
#endif
 
  out(7081h, count) 

  
  if (word_length = 16) then 
    out(7080h, startindex_ram)
  else 
    out(7080h, (startindex_ram) or shift_left(1, 31))
  endif 
     
  if (count > 0) then 
    
#IF PRIORITAET = LOW
    maxburst = 128
    if (flowrate = 1) then maxburst = 16
    if (flowrate = 3) then maxburst = 1024
         
    iindex = startindex
    icount = count 
    do 
      anzcount = min_long(icount, maxburst)
      for i = 1 to anzcount
        out(5000H, source_ar[iindex])
        inc iindex
      next
      icount = icount - anzcount
    until (icount = 0)
      
    
#else
    anzcount = count
    iindex = startindex
      
    begin_critical_section ()     
    asm{
      
      ldc source_ar  ''loads the start pointer of Data_1 in the A Register      
      ldi @___M40_SET_RAM@iindex'loads the startindex in A Register (start pointer of Data_1 will be shifted to B register)
      add   'adds A and B register      
      stindex2 (nop)   'Index e register2 points to data_1[startindex], to start from this point      
     
      ldi @___M40_SET_RAM@anzcount 'loads numvalues in A Register
     
       
      loop ma_loop_WDL2S + #Macro_call_count  'loops with "numvalues" repitions
      '------------ start of the loop
      ldi2+ (nop)  'load value to A Register and Increment index Register
      out 5000H         'Output A Register
   
         
      '-----------  'End of the loop 
      ma_loop_WDL2S + #Macro_call_count :
    }end_asm 
    end_critical_section ()

      
#endif      
                    
    do  'wait
      ready = in(7080H) 'wenn hier 1 kommt   
    until (ready = 1) 
    
  endif
       
ENDSUB

'<ENDREGION>

'<REGION> ANYBUS CC M40
#define AB_START_MSG_HEADER_OUT   800h                                          ''<HIDE>   
#define AB_START_MSG_HEADER_IN    400h                                          ''<HIDE> 
#define AB_MSG_DATA_SIZE 0                                                      ''<HIDE>   
#define AB_MSG_SOURCE_ID_OBJ 2                                                  ''<HIDE>       
#define AB_MSG_INSTANCE 3                                                       ''<HIDE> 
#define AB_MSG_COMMAND 4                                                        ''<HIDE>       
#define AB_MSG_CMD_EXT 5                                                        ''<HIDE>
#define AB_MSG_DATA 6                                                           ''<HIDE>
                      
#define AB_SOURCE_ID_OUT 03FEH                                                  ''<HIDE>  
#define AB_OBJECT_OUT 03FFH                                                     ''<HIDE>          

#define MSG_SETUP_COMPLETE 8                                                    ''<HIDE>                                 

#define ABP_MSG_HEADER_C_BIT        040h                                        ''<HIDE>                                 
#define ABP_MSG_HEADER_E_BIT        080h                                        ''<HIDE>                                 
#define ABP_MSG_HEADER_CMD_BITS     03Fh                                        ''<HIDE>                                 

#define ABP_CMD_GET_ATTR 01h              'Get attribute                        ''<HIDE>
#define ABP_CMD_SET_ATTR 02h              'Set attribute                        ''<HIDE>

#define ABP_BOOL_SIZEOF             1     'Boolean                              ''<HIDE>     
#define ABP_ANB_IA_SETUP_COMPLETE   5                                           ''<HIDE>
#define ABP_ANB_STATE_PROCESS_ACTIVE 4                                          ''<HIDE>
#define ABS_BIT_ANB_STATE 07H                                                   ''<HIDE> 

#define ABS_BIT_CMDCNT 06H                                                      ''<HIDE> 
#define ABS_BIT_CRC 0100H                                                       ''<HIDE>
#define ABS_BIT_SPI_EOF 0001H                                                   ''<HIDE>
#define ABS_BIT_SPI_CTRL_PD_VALID 001H     'Process data from ADwin is valid    ''<HIDE>
#define ABS_BIT_SPI_CTRL_CMDCNT 006H       'ADwin is fast enough to process all amounts of data           ''<HIDE>
#define ABS_BIT_SPI_CTRL_M 008H                                                 ''<HIDE>
#define ABS_BIT_SPI_CTRL_LAST 010H                                              ''<HIDE>
#define ABS_BIT_SPI_CTRL_TOGGLE 080H                                            ''<HIDE>

#define ABS_BIT_ANB_STAT_S 07H                                                  ''<HIDE>
#define ABS_BIT_SPI_STAT_M 008H                                                 ''<HIDE>
#define ABS_BIT_SPI_STAT_LAST 010H                                              ''<HIDE>


#define ABS_SPI_STATE 25                                                        ''<HIDE>
#define ABS_SPI_STATE_INIT 1                                                    ''<HIDE>
#define ABS_SPI_STATE_WRITE 2                                                   ''<HIDE>
#define ABS_SPI_STATE_START_CYCLE 3                                             ''<HIDE>    
#define ABS_SPI_STATE_CHECK_STATE 4                                             ''<HIDE>
#define ABS_SPI_STATE_READ 5                                                    ''<HIDE>
#define ABS_SPI_MSG_SIZE 26                                                     ''<HIDE>
#define ABS_SPI_PD_SIZE 27                                                      ''<HIDE>
#define ABS_SPI_HDL 28                                                          ''<HIDE>
#define ABS_SPI_CYCLE_CNT 29                                                    ''<HIDE>  
#define ABS_SPI_ERROR_CNT 30                                                    ''<HIDE>
#define ABS_LED_STAT 31                                                         ''<HIDE>
#define ABS_ANB_STAT 32                                                         ''<HIDE>
#define ABS_SPI_STAT 33                                                         ''<HIDE>
#define ABS_NETTIME 34                                                          ''<HIDE>
#define ABS_SPI_CTRL 35                                                         ''<HIDE>
#define ABS_SPI_APP_STATE 36                                                    ''<HIDE>
#define ABS_SPI_INT_MASK 37                                                     ''<HIDE>
'#define ABS_MSG_OUT_PENDING 38                                                  ''<HIDE>
#define ABS_DO_MSG 39                                                           ''<HIDE>
#define AB_ERROR_BASE 40                                                        ''<HIDE>
#define AB_ERROR_Extention1 41                                                  ''<HIDE>
#define AB_ERROR_Extention2 42                                                  ''<HIDE>
#define AB_ERROR_Extention3 43                                                  ''<HIDE>
#define AB_OUTSTANDING_COMMANDS 44                                              ''<HIDE>
#define AB_CmdState 45                                                          ''<HIDE>
#define AB_PROC_TYPE_IN 46                                                      ''<HIDE>
#define AB_PROC_CNT_IN 47                                                       ''<HIDE>
#define AB_PROC_TYPE_OUT 48                                                     ''<HIDE>  
#define AB_PROC_CNT_OUT 49                                                      ''<HIDE>
#define ABS_SPI_NEW 50                                                          ''<HIDE>  
#define AB_INSTANCE_READ_REQUEST 51                                             ''<HIDE>
#define AB_READ_SOURCE_ID 52                                                    ''<HIDE>  
#define AB_READ_OBJ_NUM 53                                                      ''<HIDE>
#define AB_READ_INST_NUM 54                                                     ''<HIDE>  
#define AB_READ_ATTR_NUM 55                                                     ''<HIDE>
#define AB_MAPPING_CNT_IN 56                                                    ''<HIDE>
#define AB_MAPPING_CNT_OUT 57                                                   ''<HIDE>
#define AB_MAPPING_SIZE_IN 58                                                   ''<HIDE>
#define AB_MAPPING_SIZE_OUT 59                                                  ''<HIDE>
#define AB_MAPPING_DATA_TYPE 60                                                 ''<HIDE>
#define AB_DATA_TYPE_SINT32 3                                                   ''<HIDE>
#define AB_DATA_TYPE_UINT32 6                                                   ''<HIDE>
#define AB_DATA_TYPE_FLOAT 18                                                   ''<HIDE>
#define AB_DIR_IN 1                                                             ''<HIDE>  
#define AB_DIR_OUT 2                                                            ''<HIDE>  
#define AB_MAPPING_DEV_ADR 61                                                   ''<HIDE>
#define ABP_BUFFER_CONTROL_IDX 63                                               ''<HIDE>
#define ABP_TEMP_IDX 64                                                         ''<HIDE>
#define ABP_STATUS_REGISTER_IDX 65                                              ''<HIDE>


#define ABP_BUFFER_CONTROL_ADR 01FFCH                                           ''<HIDE>
#define ABP_STATUS_REGISTER_ADR 01FFBH                                          ''<HIDE>
#define ABP_BIT_WRPD 01H                                                        ''<HIDE>
#define ABP_BIT_RDPD 02H                                                        ''<HIDE>
#define ABP_BIT_WRMSG 04H                                                       ''<HIDE>
#define ABP_BIT_RDMSG 08H                                                       ''<HIDE>
#define ABP_BIT_ANBR 010H                                                       ''<HIDE>
#define ABP_BIT_APPR 020H                                                       ''<HIDE>
#define ABP_PROCESS_DATA_WRITE_ADR 00000H                                       ''<HIDE>
#define ABP_PROCESS_DATA_READ_ADR 00800H                                        ''<HIDE>
#define ABP_MSG_DATA_WRITE_ADR 01000H                                           ''<HIDE>
#define ABP_MSG_DATA_READ_ADR 01800H                                            ''<HIDE>
#define ABP_MSG_DATA_LENGH 1536                                                 ''<HIDE>
'From this point the definitions are imported from the HMS C-driver example
'------------------------------------------------------------------------------
' Host application objects
'------------------------------------------------------------------------------

#define ABP_OBJ_NUM_PNIO            246   'PROFINET IO                          ''<HIDE>
#define ABP_OBJ_NUM_CCL             247   'CC-Link                              ''<HIDE>                    
#define ABP_OBJ_NUM_EIP             248   'EtherNet/IP                          ''<HIDE>
#define ABP_OBJ_NUM_ETN             249   'Ethernet                             ''<HIDE>
#define ABP_OBJ_NUM_MOD             250   'Modbus                               ''<HIDE>
#define ABP_OBJ_NUM_COP             251   'CANopen                              ''<HIDE>
#define ABP_OBJ_NUM_DEV             252   'DeviceNet                            ''<HIDE>
#define ABP_OBJ_NUM_DPV1            253   'PROFIBUS DP-V1                       ''<HIDE>
#define ABP_OBJ_NUM_APPD            254   'Application Data                     ''<HIDE>
#define ABP_OBJ_NUM_APP             255   'Application                          ''<HIDE>

'------------------------------------------------------------------------------
'State variable that keeps track of which initialization command to send next
'It is also used as a "source ID" to simplify the identification of responses
'from the Anybus-CC.
'------------------------------------------------------------------------------
#define MSG_INIT 0                                                              ''<HIDE>
#define MSG_NODE_ADR 1                                                          ''<HIDE>                   
#define MSG_BAUDRATE 2                                                          ''<HIDE>                                                          
#define MSG_MAP_IO_1 3                                                          ''<HIDE>
#define MSG_MAP_IO_2 4                                                          ''<HIDE>  
#define MSG_MAP_IO_3 5                                                          ''<HIDE>
#define MSG_MAP_IO_4 6                                                          ''<HIDE> 
#define MSG_DATA_FORMAT 7                                                       ''<HIDE>
#define MSG_SETUP_COMPLETE 8                                                    ''<HIDE>
#define MSG_DONE 9                                                              ''<HIDE> 
#define MSG_READ_REQUEST 10                                                     ''<HIDE>

'------------------------------------------------------------------------------
'Anybus module objects
'------------------------------------------------------------------------------

#define ABP_OBJ_NUM_ANB               1   'Anybus                               ''<HIDE>     
#define ABP_OBJ_NUM_DI                2   'Diagnostic                           ''<HIDE>
#define ABP_OBJ_NUM_NW                3   'Network                              ''<HIDE>
#define ABP_OBJ_NUM_NC                4   'Network Configuration                ''<HIDE>
#define ABP_OBJ_NUM_ADD               5   'PROFIBUS DP-V1 Additional Diag       ''<HIDE>
#define ABP_OBJ_NUM_RSV1              6   'Reserved                             ''<HIDE>
#define ABP_OBJ_NUM_SOC               7   'Socket Interface                     ''<HIDE>
#define ABP_OBJ_NUM_NWCCL             8   'Network CC-Link                      ''<HIDE>
#define ABP_OBJ_NUM_SMTP              9   'SMTP Client                          ''<HIDE>
#define ABP_OBJ_NUM_FSI              10   'File System Interface                ''<HIDE>
#define ABP_OBJ_NUM_NWDPV1           11   'Network PROFIBUS DP-V1               ''<HIDE>
#define ABP_OBJ_NUM_NWETN            12   'Network Ethernet                     ''<HIDE>
#define ABP_OBJ_NUM_CPC              13   'CIP Port Configuration               ''<HIDE>
#define ABP_OBJ_NUM_NWPNIO           14   'Network PROFINET IO                  ''<HIDE>
#define ABP_OBJ_NUM_PNIOADD          15   'PROFINET IO Additional Diag          ''<HIDE>

'------------------------------------------------------------------------------
'The Network object specific message commands.
'------------------------------------------------------------------------------

#define ABP_NW_CMD_MAP_ADI_WRITE_AREA        010h                               ''<HIDE>                            
#define ABP_NW_CMD_MAP_ADI_READ_AREA         011h                               ''<HIDE>
#define ABP_NW_CMD_MAP_ADI_WRITE_EXT_AREA    012h                               ''<HIDE>                            
#define ABP_NW_CMD_MAP_ADI_READ_EXT_AREA     013h                               ''<HIDE>

'The object instance number.
'------------------------------------------------------------------------------

#define ABP_INST_OBJ                0                                           ''<HIDE>

'------------------------------------------------------------------------------
'The object standard attributes.
'------------------------------------------------------------------------------

#define ABP_OA_NAME                 1     'Object name                          ''<HIDE>
#define ABP_OA_REV                  2     'Object revision                      ''<HIDE>
#define ABP_OA_NUM_INST             3     'Current number of instances          ''<HIDE>
#define ABP_OA_HIGHEST_INST         4     'Current highest instance number      ''<HIDE>

'------------------------------------------------------------------------------
'The data size of the object standard attributes (in bytes).
'------------------------------------------------------------------------------

#define ABP_OA_REV_DS               1                                           ''<HIDE>
#define ABP_OA_NUM_INST_DS          2                                           ''<HIDE>
#define ABP_OA_HIGHEST_INST_DS      2                                           ''<HIDE>
'------------------------------------------------------------------------------
'The Application data object specific message commands.
'------------------------------------------------------------------------------

#define ABP_APPD_CMD_GET_INST_BY_ORDER    010h                                  ''<HIDE>
#define ABP_APPD_CMD_GET_NEXT_GUARDED     011h                                  ''<HIDE>
#define ABP_APPD_CMD_GET_INST_NUMBERS     015h                                  ''<HIDE>
#define ABP_APPD_CMD_GET_ATTR_NAME        001h                                  ''<HIDE>
#define ABP_APPD_CMD_GET_ATTR_DATA_TYPE   002h                                  ''<HIDE>
#define ABP_APPD_CMD_GET_ATTR_NO_ELEMENTS 003h                                  ''<HIDE>
#define ABP_APPD_CMD_GET_ATTR_DISCRIPTOR  004h                                  ''<HIDE>


'Networktypes
#define AB_NETWORK_TYPE_PROFINET_IRT      089H                                  ''<HIDE>
#define AB_NETWORK_TYPE_PROFINET_FO       09DH                                  ''<HIDE>
#define AB_NETWORK_TYPE_PROFIBUS          005H                                  ''<HIDE>
#define AB_NETWORK_TYPE_ETHERCAT          087H                                  ''<HIDE>

'ADWIN ERRORCODES
#define AB_ERROR_CODE_UNSUP_OBJECT 3                                            ''<HIDE>
#define AB_ERROR_CODE_UNSUP_INSTANCE 4                                          ''<HIDE>
#define AB_ERROR_CODE_ATTR_NOT_GETTABLE 6                                       ''<HIDE>
#define AB_ERROR_CODE_UNSUP_COMMAND 5                                           ''<HIDE>
#define AB_ERROR_CODE_SETUP_FAILED 20                                           ''<HIDE>

#define AB_TRUE 1                                                               ''<HIDE>                               
#define AB_FALSE 0                                                              ''<HIDE>
#define AB_SENT 2                                                               ''<HIDE>
#define AB_HEADER_SIZE 6                                                        ''<HIDE>
#define AB_SPI_SEND_SIZE 32                                                     ''<HIDE>
'''Length of work array
#define AB_WORK_ARR_LEN 5000


'***************************************************************
'''Reads and writes process data from/to Anybus module and 
'''+handles mailbox communication to the module.
'''+return value:
'''+         0: Setup
'''+         2: Wait
'''+         4: Active
'''+     other: Error (configuration correct?)
'''module: module address (1..15)
'''out_pd_arr[]: array with output data
'''in_pd_arr[]: array for input data
'''out_pd_arr_len: number of output elements
'''in_pd_arr_len: number of input elements
'''work_arr[]: array for communication ADwin processor <-> Anybus module 
'''+ min. size AB_WORK_ARR_LEN 
'''flowrate: 
'''+         1=low
'''+         2=medium
'''+         3=fast
'''***************************************************************
FUNCTION Run_ANYBUS_CC_M40(out_pd_arr[], in_pd_arr[], out_pd_arr_len, in_pd_arr_len, work_arr[], flowrate) AS LONG  ''<HIDE>
  DIM map_cnt_in, map_cnt_out, id_counter, i AS LONG
  System_Array_Debug_Check(work_arr, 1, AB_WORK_ARR_LEN)
  M40_GET_RAM(work_arr, ABP_STATUS_REGISTER_IDX, ABP_STATUS_REGISTER_ADR, 1, flowrate, 16)
  M40_GET_RAM(work_arr, ABP_BUFFER_CONTROL_IDX, ABP_BUFFER_CONTROL_ADR, 1, flowrate, 16)   
  IF ((work_arr[ABP_BUFFER_CONTROL_IDX] AND ABP_BIT_RDMSG) = ABP_BIT_RDMSG) THEN       'There is a message available.
    M40_GET_RAM(work_arr, AB_START_MSG_HEADER_IN, ABP_MSG_DATA_READ_ADR, ABP_MSG_DATA_LENGH, flowrate, 16)   'Read Message
    work_arr[AB_START_MSG_HEADER_IN + AB_OBJECT_OUT] = SHIFT_RIGHT(work_arr[AB_START_MSG_HEADER_IN + AB_MSG_SOURCE_ID_OBJ], 8)
    work_arr[AB_START_MSG_HEADER_IN + AB_SOURCE_ID_OUT] = work_arr[AB_START_MSG_HEADER_IN + AB_MSG_SOURCE_ID_OBJ] AND 0FFH
    work_arr[ABP_TEMP_IDX] = ABP_BIT_RDMSG + ABP_BIT_APPR
    M40_SET_RAM(work_arr, ABP_TEMP_IDX, ABP_BUFFER_CONTROL_ADR, 1, flowrate, 16)                   'Message ACK
    M40_GET_RAM(work_arr, ABP_BUFFER_CONTROL_IDX, ABP_BUFFER_CONTROL_ADR, 1, flowrate, 16)   
    IF ((work_arr[AB_START_MSG_HEADER_IN + AB_MSG_COMMAND] AND ABP_MSG_HEADER_C_BIT) = ABP_MSG_HEADER_C_BIT) THEN 'Message is a command
      FOR i = 0 TO AB_HEADER_SIZE
        work_arr[AB_START_MSG_HEADER_OUT + i] = work_arr[AB_START_MSG_HEADER_IN + i]
      NEXT i
      work_arr[AB_START_MSG_HEADER_OUT + AB_SOURCE_ID_OUT] = work_arr[AB_START_MSG_HEADER_IN + AB_SOURCE_ID_OUT]
      work_arr[AB_START_MSG_HEADER_OUT + AB_OBJECT_OUT] = work_arr[AB_START_MSG_HEADER_IN + AB_OBJECT_OUT]
      work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] = work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] AND NOT(ABP_MSG_HEADER_C_BIT)
      SELECTCASE work_arr[AB_START_MSG_HEADER_IN + AB_OBJECT_OUT]
        CASE ABP_OBJ_NUM_APPD
          work_arr[ABS_DO_MSG] = AB_TRUE
          IF (ApplicationDataObject(work_arr) = AB_FALSE) THEN 
            work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] = work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] OR (ABP_MSG_HEADER_E_BIT)
            work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 1
            work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = work_arr[AB_ERROR_BASE]
          ENDIF
        CASE ABP_OBJ_NUM_APP
          work_arr[ABS_DO_MSG] = AB_TRUE
          IF (ApplicationObject(work_arr) = AB_FALSE) THEN
            work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] = work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] OR (ABP_MSG_HEADER_E_BIT)
            work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 1
            work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = work_arr[AB_ERROR_BASE]
          ENDIF
        CASEELSE
          work_arr[AB_ERROR_BASE] = AB_ERROR_CODE_UNSUP_OBJECT
          work_arr[AB_ERROR_Extention1] = work_arr[AB_START_MSG_HEADER_IN + AB_OBJECT_OUT]
          work_arr[ABS_DO_MSG] = AB_TRUE
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 1
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = work_arr[AB_ERROR_BASE]
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] = work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] OR (ABP_MSG_HEADER_E_BIT)
      ENDSELECT
    ELSE 'Message is a response to a command we have sent
      DEC work_arr[AB_OUTSTANDING_COMMANDS]
      SELECTCASE work_arr[AB_START_MSG_HEADER_IN + AB_SOURCE_ID_OUT]
        CASE MSG_SETUP_COMPLETE
          IF ((work_arr[AB_START_MSG_HEADER_IN + AB_MSG_COMMAND] AND ABP_MSG_HEADER_E_BIT) = ABP_MSG_HEADER_E_BIT) THEN
            work_arr[AB_ERROR_BASE] = AB_ERROR_CODE_SETUP_FAILED
          ENDIF
      ENDSELECT
    ENDIF
  ENDIF  ' End of message availible
  
  IF ((((work_arr[ABP_BUFFER_CONTROL_IDX] AND ABP_BIT_WRMSG) = 0) AND ((work_arr[ABP_BUFFER_CONTROL_IDX] AND ABP_BIT_ANBR) = ABP_BIT_ANBR)) AND (work_arr[ABS_DO_MSG] = AB_FALSE)) THEN 'The Anybus-CC is ready to accept commands, and the WriteMsg buffer is free
    SELECTCASE work_arr[AB_CmdState]
      CASE MSG_INIT
        map_cnt_in = 0
        map_cnt_out = 0
        id_counter = 0
        work_arr[AB_OUTSTANDING_COMMANDS] = 0
        work_arr[ABS_DO_MSG] = AB_FALSE
        work_arr[AB_CmdState] = MSG_NODE_ADR
      CASE MSG_NODE_ADR
        IF (work_arr[AB_MAPPING_DEV_ADR] <> 0) THEN
          inc id_counter
          work_arr[AB_START_MSG_HEADER_OUT + AB_SOURCE_ID_OUT] = id_counter and 0ffH
          work_arr[AB_START_MSG_HEADER_OUT + AB_OBJECT_OUT] = ABP_OBJ_NUM_NC
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_INSTANCE] = 1
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] = ABP_MSG_HEADER_C_BIT OR ABP_CMD_SET_ATTR
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 1
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_CMD_EXT] = 5                        'Node ID
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = work_arr[AB_MAPPING_DEV_ADR]
          work_arr[ABS_DO_MSG] = AB_TRUE
        ENDIF
        work_arr[AB_CmdState] = MSG_MAP_IO_1
      CASE MSG_MAP_IO_1
        inc id_counter
        work_arr[AB_START_MSG_HEADER_OUT + AB_SOURCE_ID_OUT] = id_counter and 0ffH
        work_arr[AB_START_MSG_HEADER_OUT + AB_OBJECT_OUT] = ABP_OBJ_NUM_NW
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_INSTANCE] = 1
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] = ABP_MSG_HEADER_C_BIT OR ABP_NW_CMD_MAP_ADI_WRITE_EXT_AREA
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 7
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_CMD_EXT] = 1
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = 1 + map_cnt_out
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 1] = work_arr[AB_MAPPING_SIZE_OUT]
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 2] = 0100H OR work_arr[AB_MAPPING_SIZE_OUT] 
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 3] = work_arr[AB_PROC_TYPE_OUT]
        work_arr[ABS_DO_MSG] = AB_TRUE
        INC  map_cnt_out
        IF (work_arr[AB_MAPPING_CNT_OUT] = map_cnt_out) THEN work_arr[AB_CmdState] = MSG_MAP_IO_3
      CASE MSG_MAP_IO_3
        inc id_counter
        work_arr[AB_START_MSG_HEADER_OUT + AB_SOURCE_ID_OUT] = id_counter and 0ffH
        work_arr[AB_START_MSG_HEADER_OUT + AB_OBJECT_OUT] = ABP_OBJ_NUM_NW
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_INSTANCE] = 1
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] = ABP_MSG_HEADER_C_BIT OR ABP_NW_CMD_MAP_ADI_READ_EXT_AREA
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 7
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_CMD_EXT] = 1
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = 1 + map_cnt_out + map_cnt_in 
        IF (work_arr[AB_MAPPING_DEV_ADR] <> 0) THEN work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = 6 + map_cnt_in
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 1] = work_arr[AB_MAPPING_SIZE_IN]
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 2] = 0100H OR work_arr[AB_MAPPING_SIZE_IN] 
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 3] = work_arr[AB_PROC_TYPE_IN]
        work_arr[ABS_DO_MSG] = AB_TRUE
        INC  map_cnt_in
        IF (work_arr[AB_MAPPING_CNT_IN] = map_cnt_in) THEN work_arr[AB_CmdState] = MSG_SETUP_COMPLETE
      CASE MSG_SETUP_COMPLETE
        'All the configuration commands have been processed by the Anybus - CC.
        'Go ahead and end the setup phase by sending SETUP_COMPLETE.
        inc id_counter
        work_arr[AB_START_MSG_HEADER_OUT + AB_SOURCE_ID_OUT] = id_counter and 0ffH
        work_arr[AB_START_MSG_HEADER_OUT + AB_OBJECT_OUT] = ABP_OBJ_NUM_ANB
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_INSTANCE] = 1
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] = ABP_MSG_HEADER_C_BIT OR ABP_CMD_SET_ATTR
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = ABP_BOOL_SIZEOF
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_CMD_EXT] = ABP_ANB_IA_SETUP_COMPLETE
        work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = AB_TRUE
        work_arr[ABS_DO_MSG] = AB_TRUE
        work_arr[AB_CmdState] = MSG_DONE
      CASE MSG_DONE
        IF (work_arr[AB_INSTANCE_READ_REQUEST] = AB_TRUE) THEN
          work_arr[AB_INSTANCE_READ_REQUEST] = AB_FALSE
          work_arr[AB_START_MSG_HEADER_OUT + AB_SOURCE_ID_OUT] = work_arr[AB_READ_SOURCE_ID]
          work_arr[AB_START_MSG_HEADER_OUT + AB_OBJECT_OUT] = work_arr[AB_READ_OBJ_NUM]
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_INSTANCE] = work_arr[AB_READ_INST_NUM]
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_COMMAND] = ABP_MSG_HEADER_C_BIT OR ABP_CMD_GET_ATTR
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 0
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_CMD_EXT] = work_arr[AB_READ_ATTR_NUM]
          work_arr[ABS_DO_MSG] = AB_TRUE
        ENDIF 
    ENDSELECT
    IF (work_arr[ABS_DO_MSG] = AB_TRUE) THEN
      'A new command has been posted to the Anybus - CC.
      'Increment counter for outstanding commands.
      INC work_arr[AB_OUTSTANDING_COMMANDS]
    ENDIF
  ENDIF
  IF (work_arr[ABS_DO_MSG] = AB_TRUE) THEN
    IF (Write_Msg_To_Anybus(work_arr, flowrate) = AB_TRUE) THEN
      work_arr[ABS_DO_MSG] = AB_FALSE
    endif
  ENDIF
  IF ((work_arr[ABP_STATUS_REGISTER_IDX] AND ABS_BIT_ANB_STAT_S) = ABP_ANB_STATE_PROCESS_ACTIVE) THEN
    '      'The "read process data" is only valid in the PROCESS_ACTIVE state.
    '      'Retrieve the new "read process data" from the Anybus - CC.
    IF ((work_arr[ABP_BUFFER_CONTROL_IDX] AND ABP_BIT_RDPD) = ABP_BIT_RDPD) THEN
      M40_GET_RAM(in_pd_arr, 1, ABP_PROCESS_DATA_READ_ADR, in_pd_arr_len, flowrate, 32)   
      work_arr[ABP_TEMP_IDX] = ABP_BIT_RDPD 
    ELSE
      work_arr[ABP_TEMP_IDX] = 0
    ENDIF
    
    M40_SET_RAM(out_pd_arr, 1, ABP_PROCESS_DATA_WRITE_ADR, out_pd_arr_len, flowrate, 32)   
    work_arr[ABP_TEMP_IDX] = work_arr[ABP_TEMP_IDX] + ABP_BIT_WRPD + ABP_BIT_APPR
    M40_SET_RAM(work_arr, ABP_TEMP_IDX, ABP_BUFFER_CONTROL_ADR, 1, flowrate, 16)                   'Procedata refresh
    work_arr[ABS_SPI_CTRL] = work_arr[ABS_SPI_CTRL] OR ABS_BIT_SPI_CTRL_PD_VALID
  ENDIF
  
  Run_ANYBUS_CC_M40 = work_arr[ABP_STATUS_REGISTER_IDX] AND ABS_BIT_ANB_STAT_S

ENDFUNCTION

'***************************************************************
'''Reads and writes process data from/to EtherCAT module and 
'''+handles mailbox communication to the module.
'''+return value:
'''+         0: Setup
'''+         2: Wait
'''+         3: Idle
'''+         4: Active
'''+     other: Error (configuration correct?)
'''in_pd_arr[]: array for input data
'''out_pd_arr[]: array with output data
'''in_pd_arr_len:  number (1..360) of input elements
'''out_pd_arr_len: number (1..360) of output elements
'''work_arr[]: array for communication ADwin processor <-> EtherCAT module 
'''+ min. size AB_WORK_ARR_LEN 
'''flowrate: 
'''+         1=low
'''+         2=medium
'''+         3=fast
'''***************************************************************
FUNCTION Run_EtherCAT(in_pd_arr[], out_pd_arr[], in_pd_arr_len, out_pd_arr_len, work_arr[], flowrate) AS LONG 
  Run_ETHERCAT = Run_ANYBUS_CC_M40(out_pd_arr, in_pd_arr, in_pd_arr_len, out_pd_arr_len, work_arr, flowrate)
ENDFUNCTION


'***************************************************************
'''Reads and writes process data from/to ProfinetIO module and 
'''+handles mailbox communication to the module.
'''+return value:
'''+         0: Setup
'''+         2: Wait
'''+         3: Idle
'''+         4: Active
'''+     other: Error (configuration correct?)
'''in_pd_arr[]: array for input data
'''out_pd_arr[]: array with output data
'''in_pd_arr_len:  number of in elements (1,2,4,8,16,32,64,128,196,256,320) in double words
'''out_pd_arr_len: number of out elements (1,2,4,8,16,32,64,128,196,256,320) in double words
'''work_arr[]: array for communication ADwin processor <-> ProfinetIO module 
'''+ min. size AB_WORK_ARR_LEN 
'''flowrate: 
'''+         1=low
'''+         2=medium
'''+         3=fast
'''***************************************************************
FUNCTION Run_ProfinetIO(in_pd_arr[], out_pd_arr[], in_pd_arr_len, out_pd_arr_len, work_arr[], flowrate) AS LONG
  Run_ProfinetIO = Run_ANYBUS_CC_M40(out_pd_arr, in_pd_arr, out_pd_arr_len, in_pd_arr_len, work_arr, flowrate)
ENDFUNCTION

'***************************************************************
'''Reads and writes process data from/to Profibus-40 module and 
'''+handles mailbox communication to the module.
'''+return value:
'''+         0: Setup
'''+         2: Wait
'''+         3: Idle
'''+         4: Active
'''+     other: Error (configuration correct?)
'''in_pd_arr[]: array for input data
'''out_pd_arr[]: array with output data
'''in_pd_arr_len: number of input elements
'''out_pd_arr_len: number of output elements
'''work_arr[]: array for communication ADwin processor <-> Profibus-40 module 
'''+ min. size AB_WORK_ARR_LEN 
'''flowrate: 
'''+         1=low
'''+         2=medium
'''+         3=fast
'''***************************************************************
FUNCTION Run_Profibus_M40(in_pd_arr[], out_pd_arr[], in_pd_arr_len, out_pd_arr_len, work_arr[], flowrate) AS LONG
  Run_PROFIBUS_M40 = Run_ANYBUS_CC_M40(out_pd_arr, in_pd_arr, out_pd_arr_len, in_pd_arr_len, work_arr, flowrate)
ENDFUNCTION
  
FUNCTION Write_Msg_To_Anybus(work_arr[], flowrate) AS LONG   ''<HIDE>
  dim i as long
  work_arr[ABS_SPI_MSG_SIZE] = AB_HEADER_SIZE + work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] 
  work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_SOURCE_ID_OBJ] = SHIFT_LEFT(work_arr[AB_START_MSG_HEADER_OUT + AB_OBJECT_OUT], 8) OR work_arr[AB_START_MSG_HEADER_OUT + AB_SOURCE_ID_OUT]
  M40_SET_RAM(work_arr, AB_START_MSG_HEADER_OUT, ABP_MSG_DATA_WRITE_ADR, work_arr[ABS_SPI_MSG_SIZE], flowrate, 16)                   'Message 
  work_arr[ABP_TEMP_IDX] = ABP_BIT_WRMSG + ABP_BIT_APPR
  M40_SET_RAM(work_arr, ABP_TEMP_IDX, ABP_BUFFER_CONTROL_ADR, 1, flowrate, 16)                   'Message ACK
  Write_Msg_To_Anybus = AB_TRUE
ENDFUNCTION

FUNCTION ApplicationObject(work_arr[]) AS LONG      ''<HIDE>
  ApplicationObject = AB_TRUE
  SELECTCASE work_arr[AB_START_MSG_HEADER_IN + AB_MSG_INSTANCE]
    CASE 1
      SELECTCASE work_arr[AB_START_MSG_HEADER_IN + AB_MSG_COMMAND] AND ABP_MSG_HEADER_CMD_BITS
        CASE ABP_CMD_GET_ATTR
          SELECTCASE (work_arr[AB_START_MSG_HEADER_IN + AB_MSG_CMD_EXT] AND 0FFH)
            CASEELSE
              work_arr[AB_ERROR_BASE] = AB_ERROR_CODE_ATTR_NOT_GETTABLE
              ApplicationObject = AB_FALSE
          ENDSELECT 
        CASE ABP_CMD_SET_ATTR
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 0
        CASEELSE
          work_arr[AB_ERROR_BASE] = AB_ERROR_CODE_UNSUP_COMMAND
          ApplicationObject = AB_FALSE
      ENDSELECT
    CASEELSE
      work_arr[AB_ERROR_BASE] = AB_ERROR_CODE_UNSUP_INSTANCE
      ApplicationObject = AB_FALSE
  ENDSELECT 
ENDFUNCTION 

FUNCTION ApplicationDataObject(work_arr[]) AS LONG      ''<HIDE>
  DIM i AS LONG
  ApplicationDataObject = AB_TRUE
  SELECTCASE work_arr[AB_START_MSG_HEADER_IN + AB_MSG_INSTANCE]
    CASE ABP_INST_OBJ
      SELECTCASE work_arr[AB_START_MSG_HEADER_IN + AB_MSG_COMMAND] AND ABP_MSG_HEADER_CMD_BITS
        CASE ABP_CMD_GET_ATTR
          SELECTCASE (work_arr[AB_START_MSG_HEADER_IN + AB_MSG_CMD_EXT] AND 0FFH)
            CASE ABP_OA_NAME
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = 65         '"Application data"
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 1] = 112
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 2] = 112
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 3] = 108
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 4] = 105
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 5] = 99
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 6] = 97
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 7] = 116
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 8] = 105
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 9] = 111
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 10] = 110
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 11] = 32
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 12] = 100
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 13] = 97
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 14] = 116
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 15] = 97
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 32
            CASE ABP_OA_REV
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = 1
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = ABP_OA_REV_DS
            CASE ABP_OA_NUM_INST
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = work_arr[AB_MAPPING_CNT_OUT] + work_arr[AB_MAPPING_CNT_IN]     
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = ABP_OA_NUM_INST_DS
            CASE ABP_OA_HIGHEST_INST
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = work_arr[AB_MAPPING_CNT_OUT] + work_arr[AB_MAPPING_CNT_IN]      
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = ABP_OA_HIGHEST_INST_DS
            CASE 11          'No mappabel read data
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = 0
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 2
            CASE 12          'No mappabel write data
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = 0
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 2
            CASEELSE
              work_arr[AB_ERROR_BASE] = AB_ERROR_CODE_ATTR_NOT_GETTABLE
              ApplicationDataObject = AB_FALSE
          ENDSELECT 
        CASE ABP_APPD_CMD_GET_INST_BY_ORDER
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_CMD_EXT] = work_arr[AB_START_MSG_HEADER_IN + AB_MSG_CMD_EXT]
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = work_arr[AB_START_MSG_HEADER_IN + AB_MSG_CMD_EXT] 
          work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 2
        CASE ABP_APPD_CMD_GET_INST_NUMBERS
          FOR i = 1 TO work_arr[AB_START_MSG_HEADER_IN + AB_MSG_DATA + 1]
            work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + (i - 1)] = i
          NEXT i
        CASEELSE
          work_arr[AB_ERROR_BASE] = AB_ERROR_CODE_UNSUP_COMMAND
          ApplicationDataObject = AB_FALSE
      ENDSELECT
    CASEELSE
      SELECTCASE work_arr[AB_START_MSG_HEADER_IN + AB_MSG_COMMAND] AND ABP_MSG_HEADER_CMD_BITS
        CASE ABP_CMD_GET_ATTR
          SELECTCASE (work_arr[AB_START_MSG_HEADER_IN + AB_MSG_CMD_EXT] AND 0FFH)
            case ABP_APPD_CMD_GET_ATTR_NAME
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = 050H
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 1] = 061H
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 2] = 072H
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 3] = 061H
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 4] = 06DH
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA + 5] = (030H + work_arr[AB_START_MSG_HEADER_IN + AB_MSG_INSTANCE])
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 6
            case ABP_APPD_CMD_GET_ATTR_DATA_TYPE
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = work_arr[AB_MAPPING_DATA_TYPE]
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 1
            case ABP_APPD_CMD_GET_ATTR_NO_ELEMENTS
              IF (work_arr[AB_START_MSG_HEADER_IN + AB_MSG_INSTANCE] <= work_arr[AB_MAPPING_CNT_IN]) THEN
                work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = work_arr[AB_MAPPING_SIZE_IN]
              ELSE
                work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = work_arr[AB_MAPPING_SIZE_OUT]
              ENDIF
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 1
            case ABP_APPD_CMD_GET_ATTR_DISCRIPTOR
              IF (work_arr[AB_START_MSG_HEADER_IN + AB_MSG_INSTANCE] <= work_arr[AB_MAPPING_CNT_IN]) THEN
                work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = 09H             'Get access ; Mappable as write process data
              ELSE
                work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = 013H            'Get access; Set access; Mappable as read process data
              ENDIF
              work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA_SIZE] = 1
            CASEELSE
              work_arr[AB_ERROR_BASE] = AB_ERROR_CODE_ATTR_NOT_GETTABLE
              ApplicationDataObject = AB_FALSE
          ENDSELECT 
      ENDSELECT 
  ENDSELECT 
ENDFUNCTION 
FUNCTION Check_Range_ETHERCAT(count, data_type, direction, work_arr[]) AS LONG      ''<HIDE>
  DIM size_adr AS LONG
  DIM cnt_adr AS LONG
  DIM type_size AS LONG
  IF (direction = AB_DIR_IN) THEN size_adr = AB_MAPPING_SIZE_IN
  IF (direction = AB_DIR_OUT) THEN size_adr = AB_MAPPING_SIZE_OUT
  IF (direction = AB_DIR_IN) THEN cnt_adr = AB_MAPPING_CNT_IN
  IF (direction = AB_DIR_OUT) THEN cnt_adr = AB_MAPPING_CNT_OUT
  Check_Range_ETHERCAT = AB_FALSE
  type_size = -1
  IF (data_type = AB_DATA_TYPE_UINT32) THEN type_size = 4
  IF (data_type = AB_DATA_TYPE_FLOAT) THEN type_size = 4
  IF (data_type = AB_DATA_TYPE_SINT32) THEN type_size = 4
  IF (((count * type_size) >= 0) AND ((count * type_size) <= 1440)) THEN
    Check_Range_ETHERCAT = AB_TRUE
    work_arr[AB_MAPPING_DATA_TYPE] = data_type
    work_arr[size_adr] = 001H
    work_arr[cnt_adr] = count
  ENDIF
ENDFUNCTION

FUNCTION Check_Range_PROFIBUS(dev_adr, count, direction, work_arr[]) AS LONG      ''<HIDE>
  DIM size_adr AS LONG
  DIM cnt_adr AS LONG
  DIM type_size AS LONG
  IF (direction = AB_DIR_IN) THEN size_adr = AB_MAPPING_SIZE_IN
  IF (direction = AB_DIR_OUT) THEN size_adr = AB_MAPPING_SIZE_OUT
  IF (direction = AB_DIR_IN) THEN cnt_adr = AB_MAPPING_CNT_IN
  IF (direction = AB_DIR_OUT) THEN cnt_adr = AB_MAPPING_CNT_OUT
  work_arr[AB_MAPPING_DATA_TYPE] = AB_DATA_TYPE_UINT32
  Check_Range_PROFIBUS = AB_FALSE
  work_arr[AB_MAPPING_DEV_ADR] = dev_adr
  work_arr[cnt_adr] = 1
  IF (count = 001H) THEN
    Check_Range_PROFIBUS = AB_TRUE
    work_arr[size_adr] = 001H
  ENDIF
  IF (count = 002H) THEN 
    Check_Range_PROFIBUS = AB_TRUE
    work_arr[size_adr] = 002H
  ENDIF
  IF (count = 004H) THEN 
    Check_Range_PROFIBUS = AB_TRUE
    work_arr[size_adr] = 004H
  ENDIF
  IF (count = 008H) THEN 
    Check_Range_PROFIBUS = AB_TRUE
    work_arr[size_adr] = 008H
  ENDIF
  IF (count = 010H) THEN 
    Check_Range_PROFIBUS = AB_TRUE
    work_arr[size_adr] = 010H
  ENDIF
  IF (count = 020H) THEN
    Check_Range_PROFIBUS = AB_TRUE
    work_arr[size_adr] = 020H
  ENDIF
  IF (count = 03DH) THEN
    Check_Range_PROFIBUS = AB_TRUE
    'work_arr[cnt_adr] = 2
    work_arr[size_adr] = 03DH
  ENDIF
ENDFUNCTION

FUNCTION Check_Range_PROFINET(count, direction, work_arr[]) AS LONG      ''<HIDE>
  DIM size_adr AS LONG
  DIM cnt_adr AS LONG
  IF (direction = AB_DIR_IN) THEN size_adr = AB_MAPPING_SIZE_IN
  IF (direction = AB_DIR_OUT) THEN size_adr = AB_MAPPING_SIZE_OUT
  IF (direction = AB_DIR_IN) THEN cnt_adr = AB_MAPPING_CNT_IN
  IF (direction = AB_DIR_OUT) THEN cnt_adr = AB_MAPPING_CNT_OUT
  work_arr[AB_MAPPING_DATA_TYPE] = AB_DATA_TYPE_UINT32
  Check_Range_PROFINET = AB_FALSE
  work_arr[cnt_adr] = 1
  IF (count = 001H) THEN
    Check_Range_PROFINET = AB_TRUE
    work_arr[size_adr] = 001H
  ENDIF
  IF (count = 002H) THEN 
    Check_Range_PROFINET = AB_TRUE
    work_arr[size_adr] = 002H
  ENDIF
  IF (count = 004H) THEN 
    Check_Range_PROFINET = AB_TRUE
    work_arr[size_adr] = 004H
  ENDIF
  IF (count = 008H) THEN 
    Check_Range_PROFINET = AB_TRUE
    work_arr[size_adr] = 008H
  ENDIF
  IF (count = 010H) THEN 
    Check_Range_PROFINET = AB_TRUE
    work_arr[size_adr] = 010H
  ENDIF
  IF (count = 020H) THEN
    Check_Range_PROFINET = AB_TRUE
    work_arr[size_adr] = 020H
  ENDIF
  IF (count = 040H) THEN 
    Check_Range_PROFINET = AB_TRUE
    work_arr[size_adr] = 040H
  ENDIF
  IF (count = 080H) THEN 
    Check_Range_PROFINET = AB_TRUE
    work_arr[cnt_adr] = 2
    work_arr[size_adr] = 040H
  ENDIF
  IF (count = 0C0H) THEN 
    Check_Range_PROFINET = AB_TRUE
    work_arr[cnt_adr] = 3
    work_arr[size_adr] = 040H
  ENDIF
  IF (count = 0100H) THEN 
    Check_Range_PROFINET = AB_TRUE
    work_arr[cnt_adr] = 4
    work_arr[size_adr] = 040H
  ENDIF
  IF (count = 0140H) THEN 
    Check_Range_PROFINET = AB_TRUE
    work_arr[cnt_adr] = 5
    work_arr[size_adr] = 040H
  ENDIF
ENDFUNCTION

'***************************************************************
'''Initializes datas and variables for the communication to the 
'''ProfinetIO module.
'''+return value:
'''+         0: OK
'''+         -1: Wrong no of IO data elements
'''in_mod_cnt: number of in elements {1,2,4,8,16,32,64,128,196,256,320} in double words
'''out_mod_cnt: number of out elements {1,2,4,8,16,32,64,128,196,256,320} in double words
'''work_arr[]: array for communication ADwin processor <-> ProfinetIO module 
'''+ min. size AB_WORK_ARR_LEN 
'''***************************************************************
FUNCTION Init_ProfinetIO(in_mod_cnt, out_mod_cnt, work_arr[]) AS LONG
  DIM check_result AS LONG
  check_result = Check_Range_PROFINET(in_mod_cnt, AB_DIR_IN, work_arr) AND Check_Range_PROFINET(out_mod_cnt, AB_DIR_OUT, work_arr) 
  Init_ProfinetIO = Init_ANYBUS_CC_M40(in_mod_cnt, out_mod_cnt, AB_DATA_TYPE_UINT32, check_result, work_arr) 
ENDFUNCTION
'***************************************************************
'''Initializes datas and variables for the communication to the 
'''+ETHERCAT module.
'''+return value:
'''+         0: OK
'''+         -1: Wrong IO size; the size of process data in and 
'''+             process data out must be in the range of 0 .. 1440 byte
'''in_mod_cnt: number of in elements (0..360) in double words
'''out_mod_cnt: number of out elements (0..360) in double words
'''data_type:   data type of process data in and process data out
'''+            AB_DATA_TYPE_SINT32: signed integer, 32bit
'''+            AB_DATA_TYPE_FLOAT: floating point, 32bit
'''work_arr[]: array for communication ADwin processor <-> EtherCAT module 
'''+ min. size AB_WORK_ARR_LEN 
'''***************************************************************
FUNCTION Init_EtherCAT(in_mod_cnt, out_mod_cnt, data_type, work_arr[]) AS LONG
  DIM check_result AS LONG
  check_result = Check_Range_ETHERCAT(in_mod_cnt, data_type, AB_DIR_IN, work_arr) AND Check_Range_ETHERCAT(out_mod_cnt, data_type, AB_DIR_OUT, work_arr) 
  Init_ETHERCAT = Init_ANYBUS_CC_M40(in_mod_cnt, out_mod_cnt, data_type, check_result, work_arr) 
ENDFUNCTION

'***************************************************************
'''Initializes arrays and variables for the communication to the 
'''+Profibus module.
'''+return value:
'''+         0: OK
'''+         -1: Wrong IO size; the size of process data in and 
'''in_mod_cnt: number of in elements {1,2,4,8,16,32,61}
'''out_mod_cnt: number of out elements {1,2,4,8,16,32,61}
'''work_arr[]: array for communication ADwin processor <-> Profibus module 
'''+ min. size AB_WORK_ARR_LEN 
'''***************************************************************
FUNCTION Init_Profibus_M40(dev_adr, in_mod_cnt, out_mod_cnt, work_arr[]) AS LONG
  DIM check_result AS LONG
  check_result = Check_Range_PROFIBUS(dev_adr, in_mod_cnt, AB_DIR_IN, work_arr) AND Check_Range_PROFIBUS(dev_adr, out_mod_cnt, AB_DIR_OUT, work_arr) 
  Init_PROFIBUS_M40 = Init_ANYBUS_CC_M40(in_mod_cnt, out_mod_cnt, AB_DATA_TYPE_UINT32, check_result, work_arr) 
ENDFUNCTION

'***************************************************************
'''Initializes datas and variables for the communication to the 
'''Anybus module.
'''+return value:
'''+         0: OK
'''+         -1: Wrong no of IO data elements
'''in_mod_cnt: number of in elements {1,2,4,8,16,32,64,128,196,256,320}
'''out_mod_cnt: number of out elements {1,2,4,8,16,32,64,128,196,256,320}
'''work_arr[]: array for communication ADwin processor <-> Anybus module 
'''+ min. size AB_WORK_ARR_LEN 
'''***************************************************************
FUNCTION Init_ANYBUS_CC_M40(in_mod_cnt, out_mod_cnt, data_type, check_result, work_arr[]) AS LONG     ''<HIDE>
  DIM i AS LONG
  System_Array_Debug_Check(work_arr, 1, AB_WORK_ARR_LEN)

  IF (check_result = AB_FALSE) THEN
    Init_ANYBUS_CC_M40 = -1
  ELSE
    Init_ANYBUS_CC_M40 = 0
  
    ABM40_Reset()
    FOR i = 1 TO AB_WORK_ARR_LEN
      IF ((i < AB_MAPPING_CNT_IN) OR (i > AB_MAPPING_DEV_ADR)) THEN work_arr[i] = 0
    NEXT i

    work_arr[ABS_SPI_CTRL] = ABS_BIT_CMDCNT
    work_arr[AB_PROC_TYPE_IN] = data_type
    work_arr[AB_PROC_CNT_IN] = in_mod_cnt
    work_arr[AB_PROC_TYPE_OUT] = data_type 
    work_arr[AB_PROC_CNT_OUT] = out_mod_cnt
    work_arr[AB_CmdState] = MSG_INIT
    work_arr[AB_START_MSG_HEADER_OUT + AB_MSG_DATA] = AB_TRUE
    work_arr[ABS_SPI_STATE] = ABS_SPI_STATE_INIT
    work_arr[ABS_SPI_APP_STATE] = 0
    work_arr[ABS_SPI_INT_MASK] = 0
  ENDIF
ENDFUNCTION

SUB Request_Attribute(Source_ID, Object, Instance, Attribute, work_arr[])      ''<HIDE>
  work_arr[AB_INSTANCE_READ_REQUEST] = AB_TRUE
  work_arr[AB_READ_SOURCE_ID] = Source_ID
  work_arr[AB_READ_OBJ_NUM] = Object
  work_arr[AB_READ_INST_NUM] = Instance
  work_arr[AB_READ_ATTR_NUM] = Attribute
ENDSUB
FUNCTION Read_Attribute(response_array[], work_arr[]) AS LONG      ''<HIDE>
  DIM i AS LONG
  Read_Attribute = AB_FALSE
  IF (work_arr[AB_START_MSG_HEADER_IN + AB_SOURCE_ID_OUT] = work_arr[AB_READ_SOURCE_ID]) THEN
    FOR i = 0 TO (work_arr[AB_START_MSG_HEADER_IN + AB_MSG_DATA_SIZE] / 2) 
      response_array[i + 1] = work_arr[AB_START_MSG_HEADER_IN + AB_MSG_DATA + i]
    NEXT i
    Read_Attribute = AB_TRUE
  ENDIF
ENDFUNCTION 

'<ENDREGION>



