'<TiCoBasic Header, Headerversion 001.001>
' Foldings                       = 34,99,333,368
'<Header End>
'ADBASIC-SYSTEM-FILE--V1.00--0000--0000--INC--PRO
'
'##############################################################################
'#                        RS_LIN_TiCo.inc                                     #
'#              INCLUDE-file for PRO II RS232/485/LIN/RS422                   #
'##############################################################################
'#                                                                            #
'#  28.06.2007  mf  creation                                                  #
'#  18.09.2007  mf  changed to new TiCo adresses (slow/fast adress)           #
'#  19.09.2007  ur  bugfix at new TiCo adresses                               #
'#  28.11.2007  mf  rs_set_led, lin_set_led                                   #
'#  28.11.2007  MF  RS_reset bug                                              #
'#  01.02.2008  MF  #If adwin_system                                          #
'#  03.07.2008  ur  wakeup                                                    #
'#  26.03.2009  MF  '''+                                                      #
'#  29.06.2009  MF   '''+ RS_INIT                                             #
'#  26.10.2011  MF 'RS 422 Beta                                               #
'#  07.12.2011  OL 'timing optimized                                          #
'#  04.11.2016  MF Calc_Processdelay                                          #
'#  25.01.2021  mf   Calc_TicksToNs                                           #
'#  Attention:                                                                #
'#   - Do not copy the first source line (containing                          #
'#     "'ADBASIC-SYSTEM-FILE...") to user files !                             #
'#     Otherwise the Syntax color may be not correct.                         #
'#   - This includefile requires ADbasic 4.20.04 or higher!                   #
'#                                                                            #
'##############################################################################

#If adwin_system=ADwin_ProII
#Else
#error("This include file can only be used with the ADwin ProII")
#EndIf

'<REGION> System

'***************************************************************
''' Converts a process frequency [Hz] into processor ticks (processdelay or cycle time).
''' frequency:  frequency[Hz]
'''*************************************************************
Function Calc_Processdelay(frequency) as long
  
#if prozessor=tico1
  calc_processdelay=50000000/(frequency)
#else
  calc_processdelay=100000000/(frequency)
#endif
  
endfunction


'***************************************************************
''' Converts timer ticks into ns
''' ticks: counter value ticks into ns
'''*************************************************************
Function Calc_TicksToNs(ticks) As long
  Calc_TicksToNs=(ticks)*20
endfunction


'---------------------------------------------------------------
' If a low-priority process accesses a Pro-module it must not be
' interrupted by any other process !!!
'---------------------------------------------------------------

Sub Sub_Begin_Critical_Section() ''<HIDE>
#If PRIORITAET = LOW
  Begin_Critical_Section
#EndIf
EndSub

Sub Sub_End_Critical_Section() ''<HIDE>
#If PRIORITAET = LOW
  End_Critical_Section
#EndIf
EndSub

'******************************************************************************
'''Switches the LED on or off.
'''value: 
'''+   0 off
'''+   1 on
'''****************************************************************************
Sub Set_LED(value)
  out (80804000H, value)
EndSub

'******************************************************************************
'''Checks the LED.
'''+Return:
'''+   0 off
'''+   1 on
'''**************************************************************************
Function Check_LED() AS Long
  Check_LED=in(80804000h)
EndFunction

'<ENDREGION> 

'<REGION> RSxxx (serial interface)

'******************************************************************************
'''Sets RS LEDs on/off for the RS channel.
'''channel: number of the RS interface on the RS module
'''led_col: 
'''+          0->LED off
'''+          1->LED red
'''+          2->LED green
'''+          3->LED orange
'''****************************************************************************
Sub RS_Set_LED(channel, led_col)
  out(100002FFh+(channel), led_col)
EndSub

'******************************************************************************
'''Sets a register of the seriell controller.
'''register_no: register on the serial controller 0..31
'''value: value for the selected register
'''**************************************************************************
Sub Set_RS(register_no,value)
  out((10000000h+register_no+200H),value)
  sleep(2)
  nop
  nop
EndSub

'******************************************************************************
'''Reads the value of a register of the serial controller.
'''register_no: register on the serial controller
'''**************************************************************************
Function Get_RS(register_no) AS Long
  Sub_Begin_Critical_Section
  out(0,(register_no))
  sleep(2)
  nop
  nop
  Get_RS=in(10H)
  Sub_End_Critical_Section
EndFunction

'******************************************************************
'''Initialises a serial interface.
'''channel: number of the serial interface (one ADwin-PRO module contains 2 or 4 interfaces)
'''Baudrate: Baudrate for the serial comunication
'''+ 2304000
'''+ 1152000
'''+  460800
'''+  230400
'''+  115200
'''+   57600
'''+   38400
'''+   19200
'''+    9600
'''+    4800
'''+    2400
'''+    1200
'''+     600
'''+     300
'''parity:     
'''+          0 = no parity
'''+          1 = even
'''+          2 = odd
'''data_bits: Number of data_bits for the serial comunication (5..8)
'''stop_bits:  
'''+           0 = 1 Stopbit
'''+           1 = 1,5 Stopbits by using 5 data_bits
'''+           1 = 2 Stopbits bei 6,7 by using 8 data_bits
'''handshake:   
'''+           0 = RS-232 no handshake
'''+           1 = RS-232 Hardware handshake (RTS/CTS)
'''+           2 = RS-232 Software handshake (Xon/Xoff)
'''+           3 = RS-485 no handshake
'''+           4 = RS-422 no handshake (for RS 422 Modul only)
'''+           6 = RS-422 Software handshake (Xon/Xoff) (for RS 422 Modul only)
'''****************************************************************
Sub RS_Init(channel,baudrate,parity,data_bits,stop_bits,handshake)
  Dim offset as Long                  ' offset for the selected serial interface
  Dim divisor as long                 ' value by which the external clock has to be divided to get the selected baudrate
  Dim DLL_value as Long               ' value for the DLL register
  Dim DLH_value as Long               ' value for the DLH register
  Dim LCR_value as Long               ' value for the LCR register

  offset = ((channel - 1) * 8)

  If (Handshake>2) then         'rs485 or 422
    out ((10000021H+channel),1)
  Else                          'rs232
    out ((10000021H+channel),0)
  EndIf

  divisor = (36864000) / (baudrate*16)   ' calculates the the divider for the external clock
  DLL_value = divisor and 0ffh            ' value for the DLL register
  DLH_value = shift_right((divisor and 0ff00h),8)     ' value for the DLH register

  LCR_value = (data_bits - 5)       'Break enable
  LCR_value = LCR_value or shift_left(stop_bits,2)
  If (parity = 1) then LCR_value = LCR_value or 011000b
  If (parity = 2) then LCR_value = LCR_value or 001000b    'value for the LCR register

  Set_RS(offset +  3, 080h)           ' LCR-bit7=1    Divisor latch enable
  Set_RS(offset + 0, DLL_value)       ' DLL-reg               setting baudrate
  Set_RS(offset + 1, DLH_value)       ' DLH-reg               setting baudrate
  Set_RS(offset + 3,  (0bfh))         ' LCR-reg               Enhanced feature reg enable
  Set_RS(offset + 4,  011h)           ' setting character for softwarehandshake Xon1
  Set_RS(offset + 5,  012h)           ' setting character for softwarehandshake Xon2
  Set_RS(offset + 6,  013h)           ' setting character for softwarehandshake Xoff1
  Set_RS(offset + 7,  014h)           ' setting character for softwarehandshake Xoff2

  If (handshake = 0) then
    Set_RS(offset + 2, 010h)          ' EFR-reg               no handshake(RS232) + enhanced functions enabled
  EndIf
  If (handshake = 1) then
    Set_RS(offset + 2, 0d0h)          ' EFR-reg               hardwarehandshake(RS232) + enhanced functions enabled
  EndIf
  If (handshake = 2) then
    Set_RS(offset + 2, 01ah)          ' EFR-reg               softwarehandshake(RS232) (Xon1/Xoff1) + enhanced functions enabled
  EndIf
  If (handshake = 3) then
    Set_RS(offset + 2, 010h)          ' EFR-reg               no handshake (RS485) + enhanced functions enabled
  EndIf
  
  If (handshake = 4) then
    Set_RS(offset + 2, 010h)          ' EFR-reg               no handshake(RS422) + enhanced functions enabled
    out ((10000025H+channel),2)     'setting the RTS-Pin
  EndIf
 
  If (handshake = 6) then
    Set_RS(offset + 2, 01ah)          ' EFR-reg               softwarehandshake(RS422) (Xon1/Xoff1) + enhanced functions enabled
    out ((10000025H+channel),2)     'setting the RTS-Pin
  EndIf
  
  Set_RS(offset + 3,  LCR_value)      ' LCR-reg               Divisor latch disable; 8 data_bits
  Set_RS(offset + 4 , 040h)           ' MCR reg               enable access to TCR and TLR register
  Set_RS(offset + 6 , 087h)           ' TCR reg               FIFO trigger level
  Set_RS(offset + 4 , 04h)            ' MCR reg               FIFOrdy reg enable; is necessary to get status information of the FIFO
  Set_RS(offset + 2, 00fh)            ' FCR-reg               FIFOs enable, FIFOs clear, DMA Mode1

EndSub

'***************************************************************
'''Resets the serial interface module.
'''*************************************************************
Sub RS_Reset()
  out (00000021H,0)   'reset 200 ns !!
  sleep(30)
  out (10000022H,1)   'RS 485 CH1 485
  out (10000023H,1)   'RS 485 CH2 485
  out (10000024H,1)   'RS 485 CH3 485
  out (10000025H,1)   'RS 485 CH4 485

  out (10000026H,0)   'RS 485 CH1 receive
  out (10000027H,0)   'RS 485 CH2 receive
  out (10000028H,0)   'RS 485 CH3 receive
  out (10000029H,0)   'RS 485 CH4 receive
EndSub

'***************************************************************
'''Writes a byte to the transmit FIFO, If the FIFO is not full.
'''+return value: 
'''+             0 = byte is written
'''+             1 = FIFO full
'''channel: number of the serial interface (one ADwin-PRO module contains 2 or 4 interfaces)
'''value: value which should be written to the FIFO
'''*************************************************************
Function Write_FIFO(channel,value) as Long
  Dim offset as Long                  ' offset for the selected serial interface
  Dim bit as Long

  bit = shift_left(1, channel - 1)

  offset = (channel - 1)  * 8
  If ((Get_RS(7 + offset) and bit) <> 0) then                                                       ' If the FIFO is not full then the byte is written
    Set_RS(offset, value)                                                                           ' Writeing byte to transmit FIFO
    Write_FIFO = 0                    ' return value: byte is written successful to the transmit FIFO
  Else
    Write_FIFO = 1                    ' return value: Fifo full; byte is not written
  EndIf
EndFunction

'***************************************************************
'''Checks if there is any more data to transmit.
'''+return value:
'''+ Check_Shift_Reg 
'''+               0 = Transmit Shift register and transmit hold are empty
'''+               1 = Transmit Shift register and transmit hold are not empty
'''channel: number of the serial interface (one ADwin-PRO module contains 2 or 4 interfaces)
'''***************************************************************
Function Check_Shift_Reg(channel) as Long
  Dim offset as Long                  ' offset for the selected serial interface
  Dim ret as Long                     ' return value

  offset = (channel - 1)  * 8

  ret = Get_RS(offset + 5) and 64     ' reading the status byte
  If (ret = 64) then
    Check_Shift_Reg = 0
  Else
    Check_Shift_Reg = 1
  EndIf
EndFunction

'***************************************************************
'''Reads byte from the receive FIFO.
'''+return value:
'''+  0 - 255 = value of the received byte
'''+       -1 = no bytes in the receive FIFO
'''channel: number of the serial interface (one ADwin-PRO module contains 2 or 4 interfaces)
'''*************************************************************
Function Read_FIFO(channel) as Long
  Dim offset as Long                                                                                ' offset for the selected serial interface
  offset = (channel - 1) * 8
  If ((Get_RS(offset + 5) and 1) <> 0) then                                                         ' read byte, If there is a byte in the receive FIFO
    Read_FIFO = Get_RS(offset)                                                                      ' read byte from receive FIFO (return value = value of the received byte)
  Else
    Read_FIFO = -1                                                                                  ' no new byte in the receive FIFO (return value = -1)
  EndIf
EndFunction

'***************************************************************
'''Selects whether the interface is sending data or is listening on the bus.
'''+Only for RS485.
'''channel: number of the serial interface (one ADwin-Pro module contains 2 or 4 interfaces)
'''direction: direction of communication
'''+               0 = Listening
'''+               1 = Sending Data
'''+               2 = Sending Data + Listening
'''*************************************************************
Sub RS485_Send(channel,direction)
  out ((10000025H+channel),direction)     'setting the RTS-Pin
EndSub

'<ENDREGION> 

'<REGION> LIN

'******************************************************************************
'Set LIN Master (3-0)
'1->Master
Sub Set_Master(val) ''<HIDE>
  out(0032h, val)
EndSub

'******************************************************************************
'Set LIN Wake (3-0)
Sub Set_Wake(val) ''<HIDE>
  out(0031h, val)
EndSub

'******************************************************************************
'Set LIN NSLP (3-0)
Sub Set_NSLP(val) ''<HIDE>
  out(0030h, val)
EndSub

'******************************************************************************
'''Sets LIN LEDs on/off for the LIN channel.
'''channel: number of the LIN interface on the LIN module
'''led_col: 
'''+          0->LED off
'''+          1->LED red
'''+          2->LED green
'''+          3->LED orange
'''****************************************************************************
Sub LIN_Set_LED(channel, led_col)
  out(100002FFh+(channel), led_col)
EndSub
'<ENDREGION>

'<REGION> Wake-up detection

'******************************************************************************
'Wake-up detection
'The FPGA detects a LOW-Pulse in RX-Linewith a length of 150mikrosec.
'******************************************************************************
'***********************************************
' reads the value of a register of the FPGA
'       register_no: register on the serial controller
'**********************************************
Function get_wakeup(ch) AS Long       ''<HIDE>
  get_wakeup = shift_right(in(40H), ch-1) and 1
EndFunction

Sub reset_wakeup(ch)                  ''<HIDE>
  out(0040h, shift_left(1, ch-1))
Endsub

Sub reset_wakeup_all()                ''<HIDE>
  out(0040h, 0Fh)
Endsub
'<ENDREGION>
