%{
  /* system includes */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>

#include <pwd.h>                /* getpwnam(), getpwuid() */
#include <grp.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>

#include "configfile.h"
#include "configfile_yacc.h"

//#define DEBUG
#ifdef DEBUG
#define DEBUG_LEXER printf("lexer in condition %d found ->%s<-\n", YY_START, yytext);
#else 
#define DEBUG_LEXER
#endif
void configFileerror(char *);

int lex_line_count = 1;
char string_buf[40];
int string_buf_pos;

%}
%option noyywrap
%option nounput
%option noyy_top_state
%option noinput
%option stack
%option prefix="configFile"

dec		[0-9]+
oct		"0"([0-7]{1,3})
hex		"0x"[0-9a-f]+

ipseg		[0-9]{1,3}
ipv4		{ipseg}"."{ipseg}"."{ipseg}"."{ipseg}

%x devtype _link net usb
%x ip mac dec hex oct bool string sstr
%x uid gid devconfig
%%
<*>[ \t]+                               { DEBUG_LEXER; }
<*>\n                                   { DEBUG_LEXER; lex_line_count++; }
<*>=                                    { DEBUG_LEXER; return eq;}
<INITIAL,devconfig,net,usb,_link>\[[ \t]*{hex}[ \t]*\]|\[[ \t]*{dec}[ \t]*\] {
	yy_push_state(devconfig);
	// the + 1 skips the open square brace
	configFilelval.val = strtol(yytext + 1, NULL, 0);
        if ((configFilelval.val < 1) || (configFilelval.val > 0xfff)) {
                 fprintf(stderr, "error: device number must be > 0 and < 0x1000\n"); 
                 return FLEXERR;
        }
	DEBUG_LEXER;
	return sectionhead;
}
<bool>0|false|no			{ configFilelval.val = 0; yy_pop_state(); return VAL; }
<bool>1|true|yes			{ configFilelval.val = 1;  yy_pop_state(); return VAL; }
<bool><<EOF>>				{ fprintf(stderr, "error: boolean value expected\n"); return FLEXERR; }
<oct>{oct}				{ configFilelval.val = strtol(yytext, NULL, 8); yy_pop_state(); return VAL; }
<oct><<EOF>>				{ fprintf(stderr, "error: octal value expected\n"); return FLEXERR; }

<dec>{dec}				{ DEBUG_LEXER; configFilelval.val = strtol(yytext, NULL, 10); yy_pop_state(); return VAL; }
<dec><<EOF>>				{ fprintf(stderr, "error: decimal value expected\n"); return FLEXERR; }
<hex>{hex}				{ configFilelval.val = strtol(yytext, NULL, 16); yy_pop_state(); return VAL; }
<hex><<EOF>>				{ fprintf(stderr, "error: hexadecimal value expected\n"); return FLEXERR; }

<uid,gid>{dec}				{ 
	DEBUG_LEXER;
	configFilelval.val = strtol(yytext, NULL, 10); yy_pop_state(); return VAL; }
<uid><<EOF>>				{ fprintf(stderr, "error: expected user id\n"); return FLEXERR; }
<gid><<EOF>>				{ fprintf(stderr, "error: expected group id\n"); return FLEXERR; 
// <uid,gid>.*				{ strncpy(configFilelval.str, yytext, 39); yy_pop_state(); return STR; }
}


<sstr>\"      { DEBUG_LEXER; string_buf_pos = 0; yy_push_state(string);  } // string start
<string>\"        { // string end
	DEBUG_LEXER;
	string_buf[string_buf_pos] = '\0';
	strncpy(configFilelval.str, string_buf, string_buf_pos+1);
        yy_pop_state();
	yy_pop_state();
	return STR;
}
<string>[^\"]+        {// todo: handle special cases! \n\t\\...
        char *yptr = yytext;
	DEBUG_LEXER;

        while ( *yptr )
		if (string_buf_pos > 38) {
			fprintf(stderr, "String too long!");
			return FLEXERR;
		} else {
			string_buf[string_buf_pos] = *yptr++;
			string_buf_pos++;
		}
        }


<ip>{ipv4}				{ struct in_addr addr;
					  if (! inet_aton(yytext, &addr)) {
						fprintf(stderr, "error: invalid IP \"%s\"\n", yytext);
						return FLEXERR;
					  }
					  strncpy(configFilelval.str, yytext, 16);
					  yy_pop_state();
					  return IPSTR;
					}
<ip><<EOF>>				{ fprintf(stderr, "error: IP expected\n"); return FLEXERR; }


<devconfig,_link,net,usb>uid		{ DEBUG_LEXER; yy_push_state(uid); return UID;}
<devconfig,_link,net,usb>gid		{ yy_push_state(gid); return GID; }
<devconfig,_link,net,usb>mode		{ yy_push_state(oct); return MODE; }
<devconfig,_link,net,usb>desc		{ yy_push_state(sstr); return DESC; }


<devtype>link 				{ configFilelval.val = DC_LINK; yy_push_state(_link); return DEVTYPE; }
<devtype>net 				{ configFilelval.val = DC_ETHERNET;  yy_push_state(net); return DEVTYPE; }
<devtype>usb 				{ configFilelval.val = DC_USB;  yy_push_state(usb); return DEVTYPE; }
<devtype><<EOF>>			{ fprintf(stderr, "error: device type expected\n");
					  return FLEXERR;
					}

<_link>base				{ yy_push_state(hex); return BASE; }

<usb>serial				{ yy_push_state(dec); return SERIAL; }

<net>PORT				{ DEBUG_LEXER; yy_push_state(dec); return PORT; }
<net>HOSTLINK				{ DEBUG_LEXER; yy_push_state(hex); return HOSTLINK; }
<net>TIMEOUT				{ DEBUG_LEXER; yy_push_state(dec); return TIMEOUT; }
<net>COUNT				{ DEBUG_LEXER; yy_push_state(dec); return COUNT; }
<net>IP					{ DEBUG_LEXER; yy_push_state(ip); return IP; }
<net>PW					{ DEBUG_LEXER; yy_push_state(sstr); return PW; }
<devconfig>TYPE				{ DEBUG_LEXER; yy_push_state(devtype); return TYPE; }
<*>. { DEBUG_LEXER;
	fprintf(stderr, "line %d: Unrecognized character: \"%s\"\n", lex_line_count, yytext );
	return FLEXERR;
}
%%

