
#ifndef __LIBADWIN_ERRNO_H__
#define __LIBADWIN_ERRNO_H__

#include <sys/cdefs.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 * thread-specific adwin_errno
 */
// not thread-safe version
extern const char *adwin_strerror(int errnum);
// thread-safe version
extern int adwin_strerror_r(int errnum, char *buf, size_t n);
extern void adwin_perror(const char *msg);

extern int *__adwin_errno_location(void);
#define adwin_errno             (*(__adwin_errno_location()))

#ifdef __cplusplus
}
#endif

/*
 * The error codes used in adwin32.dll (win) are rather wide-spread.
 * Here we try to merge the system errno with our adwin_errno
 * Alas the number regions overlap, that is a problem.
 * Since all the errorcodes below are positive, we define that a
 * negative adwin_errno value corresponds to the positive errno
 *  (adwin_errno < 0) <=> (adwin_errno == -errno)
 */

#include <errno.h>

#define k_ok                0
#define k_timeout_to_link       1   /* Timeout error on writing to the ADwin-system. */
#define k_timeout_from_link     2   /* Timeout error on reading from the ADwin-system. */

#define k_timeout_fast_to_link      3   /* Timeout error on fast writing to the ADwin-system. */
#define k_timeout_fast_from_link    4   /* Timeout error on fast reading from the ADwin-system. */

#define k_this_link_not_allowed     10  /* The device No. is not allowed. */
#define k_this_link_is_unknown      11  /* The device No. is not known. */


#define k_function_not_allowed      15  /* Function for this device not allowed. */

#define k_wrong_version         20  /* Incompatible versions of ADwin operating system ,\ndriver (ADwin32.DLL) and/or ADbasic binary-file. */

#define k_Data_is_to_small      100 /* The Data is too small. */
#define k_Fifo_is_to_small      101 /* The Fifo is too small or not enough values. */
#define k_Fifo_to_less_data     102 /* The Fifo has not enough values. */

#define k_File_not_found        200 /* File not found. ...-ENOENT */
#define k_File_not_binaer       202 /* The file is not an ADBasic binary-file. */
#define k_File_not_valid        203 /* The file is not valid. */
#define k_File_not_btl          204 /* The file is not a BTL. */
#define k_WrongBinFile          205     /* wrong processor or damaged. */

#define k_error_net_error       1000    /* Network error (RPC). */
#define k_error_net_binding_compose 1001    /* Network error (Bind). */
#define k_error_net_stringbinding   1002    /* Network error (String bind). */
#define k_error_net_password        1003    /* Wrong password (RPC). */

#define k_error_tcp_error       2000    /* Network error (TcpIp). */
#define k_error_tcp_timout_read     2001    /* Network timeout. */
#define k_error_tcp_password        2002    /* Wrong password. */

#define k_unknown_usb_device        3000    /* USB-device is unknown. */
#define k_unknown_device        3001    /* No such ADwin device */


#endif /* __LIBADWIN_ERRNO_H__ */
