#!/usr/bin/python
# -*- coding: utf-8 -*-
# bas_dmo2.py

import sys
from PyQt5 import QtCore, QtWidgets
import ADwin
from Ui_dialog import Ui_Form

DEVICENUMBER = 0x1

BTL = "ADwin9.btl"
PROCESS = "Bas_dmo2.t91"


class Dialog(QtWidgets.QDialog):
    def __init__(self, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_Form()
        self.ui.setupUi(self)

        self.ui.btnQuit.clicked.connect(self.quit)
        self.ui.btnStart.clicked.connect(self.start)
        self.ui.btnStop.clicked.connect(self.stop)
        self.ui.sbSetpoint.valueChanged.connect(self.setpointChanged)
        self.ui.sbGain.valueChanged.connect(self.gainChanged)

        self.adw = ADwin.ADwin(DEVICENUMBER, 1)
        btl = self.adw.ADwindir + BTL
        if sys.platform == 'win32':
            process = self.adw.ADwindir + "ADbasic\\samples_ADwin\\" + PROCESS
        else:
            process = self.adw.ADwindir + "share/examples/samples_ADwin/" + PROCESS
        self.adw.Boot(btl)
        self.adw.Load_Process(process)

    def quit(self):
        self.adw.Stop_Process(1)
        self.close()

    def start(self):
        self.adw.Start_Process(1)
        self.ui.sbSetpoint.setValue(32768)
        self.ui.sbGain.setValue(10)

    def stop(self):
        self.adw.Stop_Process(1)

    def setpointChanged(self, value):
        self.ui.leSetpoint.setText(str(value))
        self.adw.Set_Par(1, value)

    def gainChanged(self, value):
        self.ui.leGain.setText(str(value))
        self.adw.Set_Par(2, value)


app = QtWidgets.QApplication(sys.argv)
dialog = Dialog()
sys.exit(dialog.exec_())
