#! /usr/bin/python
# -*- coding: cp1252 -*-
''' Demonstrate all functions from the ADwin-Python-Module '''


#   Copyright 2017 Jger Computergesteuerte Messtechnik GmbH
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import sys
import time
import ctypes
import logging
import ADwin

RAISE_EXCEPTIONS = 1
NO_YES = ("no", "yes")

DEVICENUMBER = 0x1
PROCESSORTYPE = "11" # '9' for T9, '10' for T10, '11' for T11, '12' for T12, '121' for T12.1

if PROCESSORTYPE == "9":
    PROCESS = "example.T91"
if PROCESSORTYPE == "10":
    PROCESS = "example.TA1"
if PROCESSORTYPE == "11":
    PROCESS = "example.TB1"
if PROCESSORTYPE == "12":
    PROCESS = "example.TC1"
if PROCESSORTYPE == "121":
    PROCESS = "example.TC11"


adw = ADwin.ADwin(DEVICENUMBER, RAISE_EXCEPTIONS)

# Boot ADwin-System
print "boot ADwin-system... ",
if sys.platform == "win32":
    BTL = adw.ADwindir + "adwin" + PROCESSORTYPE + ".btl"
if sys.platform.startswith("linux"):
    BTL = "adwin" + PROCESSORTYPE + ".btl"
adw.Boot(BTL)
print "ok"

# Test_Version
print "test_version()... ",
if adw.Test_Version() == 0:
    print "ok"
else:
    print "not ok"

print "process is running:", NO_YES[adw.Process_Status(1)]
print "processdelay:", adw.Get_Processdelay(1)

# ProcessorType
print "processor_type:", adw.Processor_Type()

# Load and start process
print "load process... ",
adw.Load_Process(PROCESS)
time.sleep(0.2)
print "ok"

print "start process... ",
adw.Start_Process(1)
time.sleep(0.2)
print "ok"
print "process is running:", NO_YES[adw.Process_Status(1)]

print "workload... ",
adw.Workload() # first call with invalid value
time.sleep(0.2)
print adw.Workload(), "%"

if (PROCESSORTYPE == "9") or (PROCESSORTYPE == "10"):
    print "free memory, type pm_local:", adw.Free_Mem(1), "byte"
    print "free memory, type dm_local:", adw.Free_Mem(3), "byte"
    print "free memory, type dram_ext:", adw.Free_Mem(4), "byte"
if PROCESSORTYPE == "11":
    print "free memory, type pm_local:", adw.Free_Mem(1), "byte"
    print "free memory, type em_local:", adw.Free_Mem(2), "byte"
    print "free memory, type dm_local:", adw.Free_Mem(3), "byte"
    print "free memory, type dram_ext:", adw.Free_Mem(4), "byte"
if (PROCESSORTYPE == "12") or (PROCESSORTYPE == "121"):
    print "free memory, type cm:", adw.Free_Mem(5), "kbyte"
    print "free memory, type um:", adw.Free_Mem(6), "kbyte"


# Long-Parameter
for i in range(1, 81):
    adw.Set_Par(i, i)
print "\nGet_Par: ",  
for i in range(1, 9):
    print adw.Get_Par(i),
print adw.Get_Par(10)
print "Get_Par_All:", adw.Get_Par_All()[:10]
print "Get_Par_Block:", adw.Get_Par_Block(10, 10)[:]

# Float-Parameter
for i in range(1, 81):
    adw.Set_FPar(i, i+0.1)
print "Get_FPar: ",
for i in range(1, 9):
    print adw.Get_FPar(i),
print adw.Get_FPar(10)
print "Get_FPar_All:", adw.Get_FPar_All()[:5]
print "Get_FPar_Block:", adw.Get_FPar_Block(10, 10)[:5]

try:
    # Double-Parameter
    for i in range(1, 81):
        adw.Set_FPar_Double(i, i+0.1)
    print "Get_FPar_Double: ",
    for i in range(1, 9):
        print adw.Get_FPar_Double(i),
    print adw.Get_FPar_Double(10)
    print "Get_FPar_All_Double:", adw.Get_FPar_All_Double()[:]
    print "Get_FPar_Block_Double:", adw.Get_FPar_Block_Double(10, 10)[:]
except AttributeError:
    logging.warning("double-parameter not implemented!")


# Datas
for i in range(1, 8):
    print "Data_Type", i, adw.Data_Type(i), ", Data_Length:", adw.Data_Length(i)

# long-Data
DATATYPE = ctypes.c_int32 * 100
ARR = DATATYPE(0)
for i in range(100):
    ARR[i] = i
adw.SetData_Long(ARR, 1, 1, 100)
# also possible:
# type list: adw.SetData_Long(range(100, 200), 1, 1, 100)
# type array: adw.SetData_Long(array.array('l', range(100,200)), 1, 1, 100)
print "GetData_Long:", adw.GetData_Long(1, 10, 20)[:]

# float-Data
DATATYPE = ctypes.c_float * 100
ARR = DATATYPE(0)
for i in range(100):
    ARR[i] = i + 0.1
adw.SetData_Float(ARR, 2, 1, 100)
print "GetData_Float:", adw.GetData_Float(2, 10, 20)[:5]

if (PROCESSORTYPE == "12") or (PROCESSORTYPE == "121"):
    try:
        # double-Data
        DATATYPE = ctypes.c_double * 100
        ARR = DATATYPE(0)
        for i in range(100):
            ARR[i] = i + 0.1
        adw.SetData_Double(ARR, 6, 1, 100)
        print "GetData_Double:", adw.GetData_Double(6, 10, 20)[:5]
    except AttributeError, ADwin.ADwinError:
        logging.warning("double-data not implemented!")

# Fifo, type long
print "Fifo_Full(3):", adw.Fifo_Full(3), "- Fifo_Empty(3):", adw.Fifo_Empty(3)
print "Get some values from fifo:", adw.GetFifo_Long(3, 10)[:]
print "Fifo_Full(3):", adw.Fifo_Full(3), "- Fifo_Empty(3):", adw.Fifo_Empty(3)
print "Set some values to the fifo... ",
adw.SetFifo_Long(3, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 10)
print "ok"
print "Fifo_Full(3):", adw.Fifo_Full(3), "- Fifo_Empty(3):", adw.Fifo_Empty(3)
print "clear_fifo... ",
adw.Fifo_Clear(3)
print "ok"
print "Fifo_Full(3):", adw.Fifo_Full(3), "- Fifo_Empty(3):", adw.Fifo_Empty(3)

# Fifo, type float
print "Fifo_Full(4):", adw.Fifo_Full(4), "- Fifo_Empty(4):", adw.Fifo_Empty(4)
print "Get some values from fifo:", adw.GetFifo_Float(4, 10)[:]
print "Set some values to the fifo... ",
adw.SetFifo_Float(4, [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.0], 10)
print "ok"

# Fifo, type double
if (PROCESSORTYPE == "12") or (PROCESSORTYPE == "121"):
    try:
        print "Fifo_Full(7):", adw.Fifo_Full(7), "- Fifo_Empty(7):", adw.Fifo_Empty(7)
        print "Get some values from fifo:", adw.GetFifo_Double(7, 10)[:]
        print "Set some values to the fifo... ",
        adw.SetFifo_Double(7, [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.0], 10)
        print "ok"
    except ADwin.ADwinError:
        logging.warning("double-fifo not implemented!")

# Data <-> File
print "data2file... ",
adw.Data2File('data2.dat', 2, 1, 50000, 0)
print "ok"
print "file2data... ",
adw.File2Data('data2.dat', ADwin.ADwin.ADWIN_DATATYPE_SINGLE, 2, 1)
print "ok"

# String
STR = "The answer to this is very simple. It was a joke. \n" \
    +"It had to be a number, an ordinary, smallish number, and I chose that one.\n" \
    +"Binary representations, base thirteen, Tibetan monks are all complete nonsense.\n" \
    +"I sat at my desk, stared into the garden and thought '42 will do' I typed it out. " \
    +"End of story."
adw.SetData_String(5, STR)

print "String_Length:", adw.String_Length(5)

STR = adw.GetData_String(5, 400)
print "GetData_String: ", STR.value

print "Stop_Process... ",
adw.Stop_Process(1)
print "ok"
print "Clear_Process... ",
adw.Clear_Process(1)
print "ok"
